# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

from typing import Dict, List, Callable, NamedTuple
from itertools import chain

from .ast import AST
from .ast_node import ASTNode
from .ast_node_type import ASTNodeType


class ScopeAttributes(NamedTuple):
    statements: List[ASTNode]
    parent_node: ASTNode
    parameters: List[ASTNode] = []


def extract_scopes(node: ASTNode, ast: AST) -> List[ScopeAttributes]:
    try:
        return _scope_extractors_by_node_type[node.node_type](node, ast)
    except KeyError:
        return []


def _extract_scopes_from_assert(assert_node: ASTNode, method_ast: AST) -> List[ScopeAttributes]:
    assert assert_node.node_type == ASTNodeType.ASSERT_STATEMENT

    expression_ast = method_ast.get_subtree(assert_node.condition)
    return _find_scopes_in_expressions(expression_ast)


def _extract_scopes_from_block(block_node: ASTNode, _) -> List[ScopeAttributes]:
    assert block_node.node_type == ASTNodeType.BLOCK_STATEMENT

    return [ScopeAttributes(statements=block_node.statements, parent_node=block_node)]


def _extract_scopes_from_expression_statement(
    expression_statement: ASTNode, method_ast: AST
) -> List[ScopeAttributes]:
    assert expression_statement.node_type in {
        ASTNodeType.STATEMENT_EXPRESSION,
        ASTNodeType.RETURN_STATEMENT,
        ASTNodeType.THROW_STATEMENT,
    }

    expression_ast = method_ast.get_subtree(expression_statement.expression)
    return _find_scopes_in_expressions(expression_ast)


def _extract_scopes_from_for_cycle(for_cycle: ASTNode, method_ast: AST) -> List[ScopeAttributes]:
    assert for_cycle.node_type == ASTNodeType.FOR_STATEMENT

    control_ast = method_ast.get_subtree(for_cycle.control)
    scopes = _find_scopes_in_expressions(control_ast)
    scopes.append(
        ScopeAttributes(
            statements=_get_block_statements_list(for_cycle.body), parent_node=for_cycle
        )
    )

    return scopes


def _extract_scopes_from_if_statement(if_node: ASTNode, method_ast: AST) -> List[ScopeAttributes]:
    assert if_node.node_type == ASTNodeType.IF_STATEMENT

    condition_ast = method_ast.get_subtree(if_node.condition)
    scopes = _find_scopes_in_expressions(condition_ast)
    scopes.append(
        ScopeAttributes(
            statements=_get_block_statements_list(if_node.then_statement), parent_node=if_node
        )
    )

    while (
        if_node.else_statement is not None and
        if_node.else_statement.node_type == ASTNodeType.IF_STATEMENT
    ):
        if_node = if_node.else_statement
        condition_ast = method_ast.get_subtree(if_node.condition)
        scopes.extend(_find_scopes_in_expressions(condition_ast))
        scopes.append(
            ScopeAttributes(
                statements=_get_block_statements_list(if_node.then_statement), parent_node=if_node
            )
        )

    if if_node.else_statement is not None:
        scopes.append(
            ScopeAttributes(
                statements=_get_block_statements_list(if_node.else_statement), parent_node=if_node
            )
        )

    return scopes


def _extract_scopes_from_variable_declaration(
    variable_declaration: ASTNode, method_ast: AST
) -> List[ScopeAttributes]:
    assert variable_declaration.node_type == ASTNodeType.LOCAL_VARIABLE_DECLARATION

    return list(
        chain.from_iterable(
            _find_scopes_in_expressions(method_ast.get_subtree(declarator.initializer))
            for declarator in variable_declaration.declarators
        )
    )


def _extract_scopes_from_method(method_declaration: ASTNode, _) -> List[ScopeAttributes]:
    assert method_declaration.node_type == ASTNodeType.METHOD_DECLARATION

    return [ScopeAttributes(statements=method_declaration.body, parent_node=method_declaration)]


def _extract_scopes_from_switch_statement(
    switch_statement: ASTNode, method_ast: AST
) -> List[ScopeAttributes]:
    assert switch_statement.node_type == ASTNodeType.SWITCH_STATEMENT

    expression_ast = method_ast.get_subtree(switch_statement.expression)
    scopes = _find_scopes_in_expressions(expression_ast)

    # all case statements belong to one scope
    # thats why cases are not surrounded with curly braces
    scopes.append(
        ScopeAttributes(
            statements=list(
                chain.from_iterable(case_node.statements for case_node in switch_statement.cases)
            ),
            parent_node=switch_statement,
        )
    )

    return scopes


def _extract_scopes_from_synchronized(
    synchronized_block: ASTNode, method_ast: AST
) -> List[ScopeAttributes]:
    assert synchronized_block.node_type == ASTNodeType.SYNCHRONIZED_STATEMENT

    lock_ast = method_ast.get_subtree(synchronized_block.lock)
    scopes = _find_scopes_in_expressions(lock_ast)
    scopes.append(
        ScopeAttributes(statements=synchronized_block.block, parent_node=synchronized_block)
    )

    return scopes


def _extract_scopes_from_try_statement(try_node: ASTNode, method_ast: AST) -> List[ScopeAttributes]:
    assert try_node.node_type == ASTNodeType.TRY_STATEMENT

    scopes: List[ScopeAttributes] = []

    for resource in try_node.resources:
        initializer_ast = method_ast.get_subtree(resource.value)
        scopes.extend(_find_scopes_in_expressions(initializer_ast))

    scopes.append(ScopeAttributes(statements=try_node.block, parent_node=try_node))

    for catch in try_node.catches:
        scopes.append(
            ScopeAttributes(statements=catch.block, parent_node=try_node)
        )

    if try_node.finally_block is not None:
        scopes.append(
            ScopeAttributes(statements=try_node.finally_block, parent_node=try_node)
        )

    return scopes


def _extract_scopes_from_while_cycle(
    while_cycle: ASTNode, method_ast: AST
) -> List[ScopeAttributes]:
    assert while_cycle.node_type in {ASTNodeType.DO_STATEMENT, ASTNodeType.WHILE_STATEMENT}

    condition_ast = method_ast.get_subtree(while_cycle.condition)
    scopes = _find_scopes_in_expressions(condition_ast)
    scopes.append(
        ScopeAttributes(
            statements=_get_block_statements_list(while_cycle.body), parent_node=while_cycle
        )
    )

    return scopes


def _find_scopes_in_expressions(expression_ast: AST) -> List[ScopeAttributes]:
    """
    Finds top level lambda expressions and returns their bodies.
    Each found nested scope represented by a list of its statements. List of such list is returned.
    TO-FIX: Add support for others scopes can be found in expressions like anonymous classes.
    """

    nested_scopes_statements: List[ScopeAttributes] = []
    for nested_scope in expression_ast.get_subtrees(ASTNodeType.LAMBDA_EXPRESSION):
        lambda_declaration = nested_scope.get_root()
        nested_scopes_statements.append(
            ScopeAttributes(
                statements=lambda_declaration.body,
                parent_node=lambda_declaration,
                parameters=lambda_declaration.parameters,
            )
        )

    return nested_scopes_statements


def _get_block_statements_list(node: ASTNode) -> List[ASTNode]:
    if node.node_type == ASTNodeType.BLOCK_STATEMENT:
        return node.statements

    # there may be a single statement without curly braces
    # for consistency we wrap it in a list
    return [node]


_scope_extractors_by_node_type: Dict[
    ASTNodeType, Callable[[ASTNode, AST], List[ScopeAttributes]]
] = {
    ASTNodeType.ASSERT_STATEMENT: _extract_scopes_from_assert,
    ASTNodeType.BLOCK_STATEMENT: _extract_scopes_from_block,
    ASTNodeType.DO_STATEMENT: _extract_scopes_from_while_cycle,
    ASTNodeType.FOR_STATEMENT: _extract_scopes_from_for_cycle,
    ASTNodeType.IF_STATEMENT: _extract_scopes_from_if_statement,
    ASTNodeType.LOCAL_VARIABLE_DECLARATION: _extract_scopes_from_variable_declaration,
    ASTNodeType.METHOD_DECLARATION: _extract_scopes_from_method,
    ASTNodeType.RETURN_STATEMENT: _extract_scopes_from_expression_statement,
    ASTNodeType.STATEMENT_EXPRESSION: _extract_scopes_from_expression_statement,
    ASTNodeType.SWITCH_STATEMENT: _extract_scopes_from_switch_statement,
    ASTNodeType.SYNCHRONIZED_STATEMENT: _extract_scopes_from_synchronized,
    ASTNodeType.THROW_STATEMENT: _extract_scopes_from_expression_statement,
    ASTNodeType.TRY_STATEMENT: _extract_scopes_from_try_statement,
    ASTNodeType.WHILE_STATEMENT: _extract_scopes_from_while_cycle,
}



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

import os
import unittest
from pathlib import Path
from unittest import TestCase

from aibolit.metrics.local_methods_calls.local_methods_calls import LocalMethodsCalls


class LocalMethodsCallsTest(TestCase):
    dir_path = Path(os.path.realpath(__file__)).parent

    def test_no_local_method_calls(self):
        self.assertEqual(
            LocalMethodsCalls().value(Path(self.dir_path, 'NoLocalMethodCalls.java')),
            0,
            'Could not calculate how many local methods are called when none'
        )

    @unittest.skip('Not implemented')
    def test_local_method_calls(self):
        self.assertEqual(
            LocalMethodsCalls().value(Path(self.dir_path, 'LocalMethodCalls.java')),
            4,
            'Could not calculate how many local method are called when they exist'
        )

    @unittest.skip('Not implemented')
    def test_inner_class_local_method_calls(self):
        self.assertEqual(
            LocalMethodsCalls().value(Path(self.dir_path, 'InnerClassLocalMethodCalls.java')),
            2,
            'Could not calculate how many local methods called when they exist in inner class'
        )

    @unittest.skip('Not implemented')
    def test_double_local_method_calls(self):
        self.assertEqual(
            LocalMethodsCalls().value(Path(self.dir_path, 'DoubleLocalMethodCalls.java')),
            2,
            'Could not calculate how many local methods called when they are called more than once'
        )



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

from pathlib import Path
from unittest import TestCase

from aibolit.patterns.supermethod.supermethod import SuperMethod
from aibolit.ast_framework import AST
from aibolit.utils.ast_builder import build_ast


class SuperMethodTestCase(TestCase):
    current_directory = Path(__file__).absolute().parent

    def test_empty(self):
        filepath = self.current_directory / "Empty.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = SuperMethod()
        lines = pattern.value(ast)
        self.assertEqual(len(lines), 0)

    def test_instance_of(self):
        # It has 2 matches in anonymous class!
        filepath = self.current_directory / "Anonymous.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = SuperMethod()
        lines = pattern.value(ast)
        self.assertEqual(len(lines), 1)

    def test_instance(self):
        filepath = self.current_directory / "Simple.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = SuperMethod()
        lines = pattern.value(ast)
        self.assertEqual(len(lines), 1)

    def test_several(self):
        # It has 2 matches in anonymous class!
        filepath = self.current_directory / "Several.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = SuperMethod()
        lines = pattern.value(ast)
        self.assertEqual(len(lines), 6)

    def test_nested_class(self):
        filepath = self.current_directory / "NestedClass.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = SuperMethod()
        lines = pattern.value(ast)
        self.assertEqual(len(lines), 1)

    def test_constructor(self):
        filepath = self.current_directory / "Constructor.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = SuperMethod()
        lines = pattern.value(ast)
        self.assertEqual(len(lines), 3)

    def test_complicated_constructor(self):
        filepath = self.current_directory / "ComplicatedChainConstructor.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = SuperMethod()
        lines = pattern.value(ast)
        self.assertEqual(len(lines), 0)



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT


class ExternalMethodsCalled:
    """
        Measure the number of external methods called by the class.
        @todo #183:30min Implement external method called metric
         In the methods in the class an external method for this class can
         be called. So number of external methods called by all the methods
         in the class is suggested as a metric. Implement this metric and then
         enable tests in test_external_methods_called.py
    """

    def __init__(self):
        pass

    def value(self, filename: str):
        return 0



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

import os
from unittest import TestCase
from pathlib import Path
from aibolit.metrics.lcom4.lcom4 import LCOM4


class TestLCOM4(TestCase):
    dir_path = Path(os.path.realpath(__file__)).parent
    pattern = LCOM4()

    def test_class_with_empty_method(self):
        name = 'EmptyMethod.java'
        lcom4_val = self.pattern.value(Path(self.dir_path, name))
        self.assertEqual(lcom4_val, 2)

    def test_constructor(self):
        name = 'Constructor.java'
        lcom4_val = self.pattern.value(Path(self.dir_path, name))
        self.assertEqual(lcom4_val, 1)

    def test_scope_with_anonymous(self):
        name = 'ScopeAnonymous.java'
        lcom4_val = self.pattern.value(Path(self.dir_path, name))
        self.assertEqual(lcom4_val, 2)

    def test_simple(self):
        name = 'Simple.java'
        lcom4_val = self.pattern.value(Path(self.dir_path, name))
        self.assertEqual(lcom4_val, 1)

    def test_getter_setter(self):
        name = 'GetterSetter.java'
        lcom4_val = self.pattern.value(Path(self.dir_path, name))
        # We should ignore all setters and getters
        self.assertEqual(lcom4_val, 1)

    def test_scope(self):
        name = 'Scope.java'
        lcom4_val = self.pattern.value(Path(self.dir_path, name))
        self.assertEqual(lcom4_val, 2)

    def test_overloaded(self):
        name = 'Overloaded.java'
        lcom4_val = self.pattern.value(Path(self.dir_path, name))
        self.assertEqual(lcom4_val, 1)

    def test_class_with_chain(self):
        name = 'MethodChain.java'
        lcom4_val = self.pattern.value(Path(self.dir_path, name))
        self.assertEqual(lcom4_val, 2)

    def test_overloaded_diff(self):
        name = 'OverloadedDiffComp.java'
        lcom4_val = self.pattern.value(Path(self.dir_path, name))
        self.assertEqual(lcom4_val, 1)



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

import os
from pathlib import Path
from unittest import TestCase

from aibolit.metrics.countLeaves.numberofleaves import CountNumberOfLeaves


class TestCountLeaves(TestCase):
    dir_path = Path(os.path.realpath(__file__)).parent
    count_leaves = CountNumberOfLeaves()

    def test_simple(self):
        lines = self.count_leaves.value(Path(self.dir_path, 'simple.java'))
        self.assertEqual(lines, 1)

    def test_simple1(self):
        lines = self.count_leaves.value(Path(self.dir_path, '5.java'))
        self.assertEqual(lines, 4)

    def test1(self):
        lines = self.count_leaves.value(Path(self.dir_path, '1.java'))
        self.assertEqual(lines, 106)

    def test2(self):
        lines = self.count_leaves.value(Path(self.dir_path, '2.java'))
        self.assertEqual(lines, 30)

    def test3(self):
        lines = self.count_leaves.value(Path(self.dir_path, '3.java'))
        self.assertEqual(lines, 12)

    def test4(self):
        lines = self.count_leaves.value(Path(self.dir_path, '4.java'))
        self.assertEqual(lines, 12)

    def test5(self):
        lines = self.count_leaves.value(Path(self.dir_path, 'nested_method.java'))
        self.assertEqual(lines, 21)



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

from unittest import TestCase
from pathlib import Path

from aibolit.patterns.implements_multi.implements_multi import ImplementsMultiFinder
from aibolit.ast_framework import AST
from aibolit.utils.ast_builder import build_ast


class ImplementsMultiTestCase(TestCase):
    current_directory = Path(__file__).absolute().parent

    def test_one_class_with_types(self):
        filepath = self.current_directory / "AnimatableSplitDimensionPathValue.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = ImplementsMultiFinder()
        lines = pattern.value(ast)
        self.assertEqual(lines, [])

    def test_two_classes(self):
        filepath = self.current_directory / "AnimatableTransform.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = ImplementsMultiFinder()
        lines = pattern.value(ast)
        self.assertEqual(lines, [15])

    def test_implements_in_string(self):
        filepath = self.current_directory / "AuditEventModelProcessor.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = ImplementsMultiFinder()
        lines = pattern.value(ast)
        self.assertEqual(lines, [])

    def test_implements_with_parantheses(self):
        filepath = self.current_directory / "BaseKeyframeAnimation.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = ImplementsMultiFinder()
        lines = pattern.value(ast)
        self.assertEqual(lines, [])

    def test_implements_with_nested_parantheses(self):
        filepath = self.current_directory / "Configuration.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = ImplementsMultiFinder()
        lines = pattern.value(ast)
        self.assertEqual(lines, [228])

    def test_implements_multi_classes(self):
        filepath = self.current_directory / "FillContent.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = ImplementsMultiFinder()
        lines = pattern.value(ast)
        self.assertEqual(lines, [32])

    def test_implements_with_parantheses_multi(self):
        filepath = self.current_directory / "FJIterateTest.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = ImplementsMultiFinder()
        lines = pattern.value(ast)
        self.assertEqual(lines, [597])

    def test_implements_with_parantheses_before(self):
        filepath = self.current_directory / "FJListProcedureRunner.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = ImplementsMultiFinder()
        lines = pattern.value(ast)
        self.assertEqual(lines, [])

    def test_implements_in_comments(self):
        filepath = self.current_directory / "KeyProviderCryptoExtension.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = ImplementsMultiFinder()
        lines = pattern.value(ast)
        self.assertEqual(lines, [])

    def test_implements_multi(self):
        filepath = self.current_directory / "OsSecureRandom.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = ImplementsMultiFinder()
        lines = pattern.value(ast)
        self.assertEqual(lines, [45])

    def test_implements_three(self):
        filepath = self.current_directory / "RectangleContent.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = ImplementsMultiFinder()
        lines = pattern.value(ast)
        self.assertEqual(lines, [25])

    def test_implements_many(self):
        filepath = self.current_directory / "SequenceFile.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = ImplementsMultiFinder()
        lines = pattern.value(ast)
        self.assertEqual(lines, [840])



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

import os
from pathlib import Path
from unittest import TestCase

from aibolit.metrics.spaces.SpaceCounter import IndentationCounter


class TestSpaces(TestCase):
    dir_path = Path(os.path.realpath(__file__)).parent
    max_right = IndentationCounter(max_right=True)
    max_left = IndentationCounter(max_left=True)
    left_var = IndentationCounter(left_var=True)
    right_var = IndentationCounter(right_var=True)

    def test_class_with_best_ident(self):
        val = self.max_left.value(Path(self.dir_path, 'BestIdent.java'))
        self.assertEqual(val, 44)
        val = self.max_right.value(Path(self.dir_path, 'BestIdent.java'))
        self.assertEqual(val, 113)

    def test_class_without_left_spaces(self):
        val = self.max_left.value(Path(self.dir_path, 'NoLeftSpaces.java'))
        self.assertEqual(val, 0)
        val = self.max_right.value(Path(self.dir_path, 'NoLeftSpaces.java'))
        self.assertEqual(val, 55)

    def test_class_without_right_spaces(self):
        val = self.max_left.value(Path(self.dir_path, 'NoRightSpaces.java'))
        self.assertEqual(val, 57)
        val = self.max_right.value(Path(self.dir_path, 'NoRightSpaces.java'))
        self.assertEqual(val, 0)

    def test_class_with_equal_spaces_number(self):
        val = self.max_left.value(Path(self.dir_path, 'SameMean.java'))
        self.assertEqual(val, 4)
        val = self.max_right.value(Path(self.dir_path, 'SameMean.java'))
        self.assertEqual(val, 55)

    def test_class_with_tabs_and_spaces(self):
        val = self.max_left.value(Path(self.dir_path, 'SpacesAndTabs.java'))
        self.assertEqual(val, 8)
        val = self.max_right.value(Path(self.dir_path, 'SpacesAndTabs.java'))
        self.assertEqual(val, 59)

    def test_class_with_worst_ident(self):
        val = self.max_left.value(Path(self.dir_path, 'WorstIdentation.java'))
        self.assertEqual(val, 20)
        val = self.max_right.value(Path(self.dir_path, 'WorstIdentation.java'))
        self.assertEqual(val, 163)

    def test_empty_examples(self):
        val = self.max_left.value(Path(self.dir_path, 'ClusterDataSourceConfiguration.java'))
        self.assertEqual(val, 0)
        val = self.max_right.value(Path(self.dir_path, 'ClusterDataSourceConfiguration.java'))
        self.assertEqual(val, 0)
        val = self.left_var.value(Path(self.dir_path, 'ClusterDataSourceConfiguration.java'))
        self.assertEqual(val, 0)
        val = self.right_var.value(Path(self.dir_path, 'ClusterDataSourceConfiguration.java'))
        self.assertEqual(val, 0)

    def test_one_point_in_series(self):
        val = self.max_left.value(Path(self.dir_path, 'package-info.java'))
        self.assertEqual(val, 0)
        val = self.max_right.value(Path(self.dir_path, 'package-info.java'))
        self.assertEqual(val, 0)
        val = self.left_var.value(Path(self.dir_path, 'package-info.java'))
        self.assertEqual(val, 0)
        val = self.right_var.value(Path(self.dir_path, 'package-info.java'))
        self.assertEqual(val, 0)



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

import os
from unittest import TestCase
from aibolit.metrics.loc.loc import Loc


class LocTest(TestCase):
    def test_it_works(self):
        m = Loc(os.path.dirname(os.path.realpath(__file__)) + '/sample-1.java')
        loc = m.value()
        self.assertEqual(loc, 8)



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT
from typing import List

from aibolit.ast_framework import ASTNodeType, AST
from aibolit.ast_framework.ast_node import ASTNode


class NullCheck():
    '''
    If we check that something equals
    (or not equals) null (except in constructor)
    it is considered a pattern.
    '''
    def value(self, ast: AST) -> List[int]:
        lines: List[int] = []
        for method_declaration in ast.get_proxy_nodes(ASTNodeType.METHOD_DECLARATION):
            for bin_op in ast.get_subtree(method_declaration).get_proxy_nodes(
                    ASTNodeType.BINARY_OPERATION):
                if self._check_null(bin_op):
                    lines.append(bin_op.operandr.line)
        return lines

    def _check_null(self, bin_operation: ASTNode) -> bool:
        return (bin_operation.operator in ["==", "!="] and
                bin_operation.operandr.node_type == ASTNodeType.LITERAL and
                bin_operation.operandr.value == "null")



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

from typing import List

from aibolit.ast_framework import AST, ASTNodeType


class AssertInCode:
    '''
    Finds use of all asserts.
    '''

    def value(self, ast: AST) -> List[int]:
        lines: List[int] = []
        for assert_node in ast.get_proxy_nodes(ASTNodeType.ASSERT_STATEMENT):
            lines.append(assert_node.line)

        return lines



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

from pathlib import Path
from unittest import TestCase

from aibolit.patterns.send_null.send_null import SendNull
from aibolit.ast_framework import AST
from aibolit.utils.ast_builder import build_ast


class SendNullTestCase(TestCase):
    current_directory = Path(__file__).absolute().parent

    def test_one_send(self):
        filepath = self.current_directory / "BaseKeyframeAnimation.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = SendNull()
        lines = pattern.value(ast)
        self.assertEqual(lines, [149])

    def test_multi_level_invocation(self):
        filepath = self.current_directory / "Configuration.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = SendNull()
        lines = pattern.value(ast)
        self.assertEqual(
            lines,
            [382, 445, 552, 641, 659, 833, 869, 1365, 2396, 2877, 2991, 3083, 3495, 3761, 3858]
        )

    def test_no_null_methods(self):
        filepath = self.current_directory / "FillContent.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = SendNull()
        lines = pattern.value(ast)
        self.assertEqual(lines, [])

    def test_simple_invocation(self):
        filepath = self.current_directory / "FJIterateTest.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = SendNull()
        lines = pattern.value(ast)
        self.assertEqual(lines, [489])

    def test_more_method_invocations(self):
        filepath = self.current_directory / "SequenceFile.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = SendNull()
        lines = pattern.value(ast)
        self.assertEqual(lines, [1100, 1189, 1204, 1220, 3288, 3301, 3370, 3540, 3553])

    def test_constructor_send_null(self):
        filepath = self.current_directory / "Constructor.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = SendNull()
        lines = pattern.value(ast)
        self.assertEqual(lines, [8, 17, 18, 19, 20, 21])

    def test_super_in_constructor_with_ternary_operator(self):
        filepath = self.current_directory / "AclPermissionParam.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = SendNull()
        lines = pattern.value(ast)
        self.assertEqual(lines, [49, 53])

    def test_this_with_ternary_operator(self):
        filepath = self.current_directory / "AddOp.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = SendNull()
        lines = pattern.value(ast)
        self.assertEqual(lines, [31, 35])

    def test_super_in_constructor_with_method_inv(self):
        filepath = self.current_directory / "ByteArrayMultipartFileEditor.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = SendNull()
        lines = pattern.value(ast)
        self.assertEqual(lines, [51])



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT
from typing import Set, List

from aibolit.ast_framework import ASTNodeType, AST
from aibolit.ast_framework.ast_node import ASTNode


class StringConcatFinder:
    '''
    Any usage string concatenation using '+' operator is considered as a pattern.
    '''
    def _check_left_right_operator(self, node: ASTNode) -> bool:
        assert node.node_type == ASTNodeType.BINARY_OPERATION
        for operator_side in [node.operandr, node.operandl]:
            if (operator_side.node_type == ASTNodeType.LITERAL and
                    isinstance(operator_side.value, str) and
                    not operator_side.value.isdigit()):
                return True
        return False

    def value(self, ast: AST) -> List[int]:
        lines: Set[int] = set()
        for node in ast.get_proxy_nodes(ASTNodeType.BINARY_OPERATION):
            if node.operator == '+' and self._check_left_right_operator(node):
                lines.add(node.line)

        return sorted(lines)



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT
import os
import shutil
import subprocess
from pathlib import Path
import pickle
import pandas as pd  # type: ignore

from aibolit.model.model import PatternRankingModel, scale_dataset  # type: ignore
from aibolit.config import Config


def collect_dataset(args):
    """
    Run bash scripts to collect metrics and patterns for java files
    """

    def make_patterns(args, cur_work_dir):
        print('Compute patterns...')
        result = subprocess.run(['make', 'patterns'], stdout=subprocess.PIPE,
                                encoding='utf-8', cwd=cur_work_dir, check=False)
        print(result.returncode)
        if result.returncode != 0:
            print(result.stderr)
            exit(3)
        else:
            print(result.stdout)
            if args.dataset_file:
                dataset_file_path = Path(cur_work_dir, args.dataset_file)
                if not dataset_file_path.parent.exists():
                    dataset_file_path.parent.mkdir(parents=True)
                shutil.copy(Path(Config.dataset_file()), dataset_file_path)
            else:
                dataset_file_path = Path(Config.dataset_file())
            print(f'dataset was saved to {str(dataset_file_path.absolute())}')

    def run_cmd(metrics_cmd, cur_work_dir):
        result = subprocess.run(metrics_cmd, stdout=subprocess.PIPE,
                                encoding='utf-8', cwd=cur_work_dir, check=False)
        if result.returncode != 0:
            print(result.stderr)
            exit(1)
        else:
            print(result.stdout)

    # path to java files which will be analyzed
    java_folder = args.java_folder
    max_classes = args.max_classes
    os.chdir(Path(Config.home_aibolit_folder(), 'scripts'))
    if not java_folder:
        java_folder = Path('target/01').absolute()
        print(f'Analyzing {java_folder} dir:')
    cur_work_dir = Path(os.getcwd())
    print('Current working directory: ', cur_work_dir)
    print('Directory with JAVA classes: ', java_folder)
    print('Filtering java files...')

    filter_cmd = ['make', 'filter']
    merge_cmd = ['make', 'merge']
    split_cmd = ['make', 'split']

    if java_folder is not None:
        filter_cmd.append(f'dir={java_folder}')
    if max_classes is not None:
        filter_cmd.append(f'max_classes={max_classes}')

    run_cmd(filter_cmd, cur_work_dir)
    make_patterns(args, cur_work_dir)

    print('Merge results...')
    run_cmd(merge_cmd, cur_work_dir)

    print('Preprocess dataset, create train and test...')
    run_cmd(split_cmd, cur_work_dir)


def train_process(target_metric_code="M4"):
    """
    Define needed columns for dataset and run model training
    """
    config = Config.get_patterns_config()
    only_patterns = [
        x['code'] for x in list(config['patterns'])
        if x['code'] not in config['patterns_exclude']
    ]
    only_metrics = \
        [x['code'] for x in list(config['metrics'])
         if x['code'] not in config['metrics_exclude']] \
        + ['halstead volume']
    columns_features = only_metrics + only_patterns
    features_number = len(columns_features)
    print("General number of features in config: ", features_number)

    train_dataset = pd.read_csv(Config.train_csv(), index_col=None)
    model = PatternRankingModel()
    # At the moment we use use patterns as features,
    # but in future features can be also some metrics.
    # We should differ them for any purpose (scaling, etc.)
    features_conf = {
        "features_order": only_patterns,
        "patterns_only": only_patterns
    }
    model.features_conf = features_conf
    print('Scaling features...')
    scaled_dataset = scale_dataset(train_dataset, model.features_conf, target_metric_code)
    dataset = scaled_dataset[only_patterns]
    print('Training model...')
    model.fit_regressor(dataset, scaled_dataset[target_metric_code])

    save_model_file = Path(Config.folder_to_save_model_data(), 'model.pkl')
    print(f'Saving model to loaded model from file {save_model_file}:')
    with open(save_model_file, 'wb') as fid:
        pickle.dump(model, fid)

    load_model_file = Path(Config.folder_to_save_model_data(), 'model.pkl')
    print(f'Test loaded model from file {load_model_file}:')
    test_dataset = pd.read_csv(Config.test_csv(), index_col=None)
    with open(load_model_file, 'rb') as fid:
        model_new = pickle.load(fid)
        scaled_test_dataset = scale_dataset(
            test_dataset,
            model_new.features_conf,
            target_metric_code
        ).sample(n=10, random_state=17)
        print('Model has been loaded successfully')
        # add ncss, ncss is needed in informative as a  last column
        X_test = scaled_test_dataset[only_patterns + ['M2']]

        for _, row in X_test.iterrows():
            preds, importances = model_new.rank(row.values)
            print(preds)
    path_with_logs = Path(os.getcwd(), 'catboost_info')
    print(f'Removing path with catboost logs {path_with_logs}')
    if path_with_logs.exists():
        shutil.rmtree(path_with_logs)



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT
import os
import shutil
from pathlib import Path
from time import time

from aibolit.config import Config
from aibolit.model.model import PatternRankingModel, scale_dataset, generate_fake_dataset


def test_model_training():
    cur_file_dir = Path(os.path.realpath(__file__)).parent
    config = Config.get_patterns_config()
    model = PatternRankingModel()
    patterns = [x['code'] for x in config['patterns']]
    train_df = generate_fake_dataset()
    model.features_conf = {'features_order': patterns}
    scaled_df = scale_dataset(train_df, model.features_conf, "M4")
    start = time()
    print('Start training...')
    model.fit_regressor(scaled_df[patterns], scaled_df['M4'])
    end = time()
    print(f'End training. Elapsed time: {end - start:.2f} secs')
    # this folder is created by catboost library, impossible to get rid of it
    catboost_folder = Path(cur_file_dir, 'catboost_info')
    if catboost_folder.exists():
        shutil.rmtree(catboost_folder)


def test_train_with_selected_features():
    cur_file_dir = Path(os.path.realpath(__file__)).parent
    model = PatternRankingModel()
    selected_patterns = ['P18', 'P10', 'M2', 'M5']
    train_df = generate_fake_dataset()
    print(f'Features for the whole dataset: {list(train_df.columns)}')
    target = train_df.pop('M4')
    start = time()
    print('Start training...')
    model.fit_regressor(train_df, target, selected_patterns)
    end = time()
    print(f'End training. Elapsed time: {end - start:.2f} secs')
    # this folder is created by catboost library, impossible to get rid of it
    catboost_folder = Path(cur_file_dir, 'catboost_info')
    if catboost_folder.exists():
        shutil.rmtree(catboost_folder)
    print(f'Model features: {list(model.features_conf["features_order"])}')


if __name__ == '__main__':
    test_model_training()
    test_train_with_selected_features()



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT
from decimal import localcontext, ROUND_DOWN, Decimal
from typing import Dict, Any, Tuple, List, Union
from numpy.typing import NDArray

import numpy as np
import pandas as pd  # type: ignore[import-untyped]
from catboost import CatBoost  # type: ignore[import-untyped]
from sklearn.base import BaseEstimator  # type: ignore[import-untyped]

from aibolit.config import Config


def get_minimum(
        c1: NDArray[np.float64],
        c2: NDArray[np.float64],
        c3: NDArray[np.float64]) -> Tuple[NDArray[np.float64], NDArray[np.int64]]:
    """
    Args:
        c1, c2, c3: np.array with shape (number of snippets, ).
    Returns:
        c: np.array with shape (number of snippets, ) -
        elemental minimum of 3 arrays.
        number: np.array with shape (number of snippets, ) of
        arrays' numbers with minimum elements.            .
    """

    c: NDArray[np.float64] = np.vstack((c1, c2, c3))

    return np.min(c, 0), np.argmin(c, 0)


def generate_fake_dataset() -> pd.DataFrame:
    config = Config.get_patterns_config()
    patterns = [x['code'] for x in config['patterns']]
    metrics = [x['code'] for x in config['metrics']]

    train_df = pd.DataFrame(columns=patterns)
    min_rows_for_train = 10
    for x in range(min_rows_for_train):
        p = {p: (x + i) for i, p in enumerate(patterns)}
        m = {p: (x + i) for i, p in enumerate(metrics)}
        row = {**p, **m}
        train_df = pd.concat([train_df, pd.DataFrame([row])], ignore_index=True)

    train_df = train_df.astype(float)
    return train_df


def scale_dataset(
        df: pd.DataFrame,
        features_conf: Dict[Any, Any],
        target_metric_code: str,
        scale_ncss=True) -> pd.DataFrame:
    config = Config.get_patterns_config()
    patterns_codes_set = set([x['code'] for x in config['patterns']])
    metrics_codes_set = set([x['code'] for x in config['metrics']])
    exclude_features = set(config['patterns_exclude']).union(set(config['metrics_exclude']))
    used_codes = set(features_conf['features_order'])
    used_codes.add(target_metric_code)
    not_scaled_codes = set(patterns_codes_set).union(set(metrics_codes_set)).difference(used_codes).difference(
        exclude_features)
    features_not_in_config: List[str] = list(set(df.columns).difference(not_scaled_codes).difference(used_codes))
    not_scaled_codes_list: List[str] = sorted(list(not_scaled_codes.union(set(features_not_in_config))))
    codes_to_scale = sorted(used_codes)
    if scale_ncss:
        scaled_df = pd.DataFrame(
            df[codes_to_scale].values / df['M2'].values.reshape((-1, 1)),
            columns=codes_to_scale
        )
        not_scaled_df = df[not_scaled_codes_list]
        input = pd.concat([scaled_df, not_scaled_df], axis=1)
    else:
        input = df

    return input


class PatternRankingModel(BaseEstimator):

    def __init__(self):
        self.do_rename_columns = False
        self.model = None
        self.features_conf = None

    def predict(self, input_params: Dict[Any, Any]) -> Tuple[Dict[Any, int], List[float]]:
        features_order = self.features_conf['features_order']
        # add ncss to last column. We will normalize all patterns by that value
        input = [input_params[i] for i in features_order] + [input_params['M2']]
        preds, importances = self.rank(input)

        return {features_order[int(x)]: int(x) for x in preds}, list(importances)

    def fit_regressor(self, X, y, features=None, display=False) -> None:
        """

        :param X: X np.array with shape (number of snippets, number of patterns) or
                (number of patterns, ).
        :param y: np.array with shape (number of snippets,), array of snippets'
                complexity metric values
        :param features: set of features to train
        :param display: show additional output
        :return: None
        """
        model = CatBoost()

        grid = {'learning_rate': [0.03, 0.1],
                'depth': [4, 6, 10],
                'l2_leaf_reg': [1, 3, 5, 7, 9]}
        if features:
            X = X[features]

        self.features_conf = {'features_order': X.columns}

        model.grid_search(
            grid,
            X=X,
            y=y,
            verbose=display,
        )

        self.model = model
        self.model.fit(X, y.ravel(), logging_level='Silent')

    def sigmoid(self, x):
        return 1 / (1 + np.exp(-x))

    def __get_pairs(self, item, th: float, feature_importances=None):
        if not feature_importances:
            feature_importances = self.model.feature_importances_
        pattern_importances = item * feature_importances
        # mask discards not significant patterns
        th_mask = (self.sigmoid(pattern_importances) <= th) + 0
        pattern_importances = pattern_importances * th_mask
        order = np.arange(self.model.feature_importances_.size)
        return (pattern_importances, order)

    def __vstack_arrays(self, res):
        return np.vstack(res).T

    def calculate_score(self, X, quantity_func='log', th=1.0, feature_importances=None):
        """
        Args:
            X: np.array with shape ( number of patterns).
            quantity_func: str, type of function that will be applied to
                number of occurrences.
            th (float): Sensitivity of algorithm to recommend.
                0 - ignore all recommendations
                1 - use all recommendations

        Returns:
            ranked: np.array with shape (number of snippets, number of patterns)
                of sorted patterns in non-increasing order for each snippet of
                code.
        """

        X = X.copy()
        X = np.expand_dims(X, axis=0)

        ranked = []
        quantity_funcs = {
            'log': lambda x: np.log1p(x) / np.log(10),
            'exp': lambda x: np.exp(x + 1),
            'linear': lambda x: x,
        }

        try:
            item = quantity_funcs[quantity_func](X)
            pairs = self.__vstack_arrays(self.__get_pairs(item, th, feature_importances))
            pairs = pairs[pairs[:, 0].argsort()]
            ranked.append(pairs[:, 1].T.tolist()[::-1])
        except Exception:
            import traceback
            traceback.print_exc()
            raise Exception("Unknown func")

        return (np.array(ranked), pairs[:, 0].T.tolist()[::-1])

    def test(self, files: List[str]) -> List[List[Union[str, List[str], List[float]]]]:
        """Make predict for list of java files using current model."""

        config = Config.get_patterns_config()
        patterns_config = config['patterns']
        metrics_config = config['metrics']
        patterns_codes = [x['code'] for x in patterns_config]
        metrics_codes = [x['code'] for x in metrics_config]
        features = self.features_conf['features_order']
        results = []
        for file in files:
            row = {}
            for feature in features:
                # we need for test to scale snippet
                # we will calculate it if we do not have M2 in feature_conf
                found_feature = [x for x in metrics_config if x['code'] == 'M2']
                if feature == 'M2':
                    continue
                row['filename'] = file
                ncss_val = found_feature[0]['make']().value(file)
                row['M2'] = ncss_val

                if feature in patterns_codes:
                    found_feature = [x for x in patterns_config if x['code'] == feature]
                    lines = found_feature[0]['make']().value(file)
                    row[feature] = float(len(lines))  # type: ignore[assignment]
                    results.append(row)
                elif feature in metrics_codes:
                    found_feature = [x for x in metrics_config if x['code'] == feature]
                    val = found_feature[0]['make']().value(file)
                    if val:
                        row[feature] = val
                    results.append(row)
                else:
                    continue

        result_array = []
        for file_for_file in results:
            sorted_result, importances = self.predict(file_for_file)
            result_array.append([file_for_file['filename'], list(sorted_result.keys()), importances])

        return result_array  # type: ignore[return-value]

    def rank(self, snippet, scale=True):
        """
        Args:
            snippet: np.array with shape (number of snippets, number of patterns + 1),
            because last column is ncss
        Returns:
            ranked: np.array with shape (number of snippets, number of patterns)
                of sorted patterns in non-increasing order for each snippet of
                code.
        """

        # remember it, since we will use `log` function for non-normalized input value
        patterns_orig = np.array(snippet[:-1])
        ncss = snippet[-1]

        if scale:
            snippet = patterns_orig / ncss
        else:
            snippet = patterns_orig

        k = snippet.size
        complexity = self.model.predict(snippet)
        importances = []
        for i in range(k):
            if snippet[i] == 0:
                # do not need to predict if we have 0
                importances.append((i, 0))
                continue
            temp_arr = snippet.copy()
            temp_arr[i] = temp_arr[i] - (1 / ncss)
            complexity_minus = self.model.predict(temp_arr)
            if complexity_minus < complexity:
                # complexity decreased
                with localcontext() as ctx:
                    ctx.rounding = ROUND_DOWN
                    abs_diff = abs(complexity - complexity_minus)
                    diff = Decimal(abs_diff).quantize(Decimal('0.001'))
                    diff = float(diff * 100)
            else:
                # complexity increased, we do not need such value, set to 0,
                # cause we need only patterns when complexity decreased
                diff = 0
            importances.append((i, diff))

        sorted_importances = dict(sorted(importances, key=lambda x: x[1], reverse=True))
        return sorted_importances.keys(), sorted_importances.values()



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT
from typing import List

from aibolit.ast_framework import ASTNodeType, AST


class NonFinalAttribute:
    '''
    return lines of non-final attributes
    '''
    def value(self, ast: AST) -> List[int]:
        lines: List[int] = []
        for field in ast.get_proxy_nodes(ASTNodeType.FIELD_DECLARATION):
            if 'final' not in field.modifiers:
                lines.append(field.line)
        return lines



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

from typing import Callable, Iterator

from aibolit.ast_framework import ASTNode, ASTNodeType


def nodes_filter_factory(
    base_field_name: str, *node_types: ASTNodeType
) -> Callable[[ASTNode], Iterator[ASTNode]]:
    """
    Create filter, which takes 'body_field_name' field of incoming node,
    checks if it list of ASTNode, and return it filtered by node_type.
    """

    def filter(base_node: ASTNode) -> Iterator[ASTNode]:
        base_field = getattr(base_node, base_field_name)
        if isinstance(base_field, list):
            for node in base_field:
                if isinstance(node, ASTNode) and node.node_type in node_types:
                    yield node
        else:
            raise RuntimeError(
                f"Failed computing ASTNode field based on {base_field_name} field. "
                f"Expected list, but got {base_field} of type {type(base_field)}."
            )

    return filter



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT


class LocalMethodsCalls:
    """
        Measure the number of local methods called by the class.
        @todo #183:30min Implement local method called metric
         In the methods in the class an local method (method that is defined in this class)
         can be called. So number of local methods calls by all the methods in the
         class is suggested as a metric. Implement this metric and then
         enable tests in test_local_methods_calls.py
    """

    def __init__(self):
        pass

    def value(self, filename: str):
        return 0



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

import os
from unittest import TestCase
import unittest
from pathlib import Path
from aibolit.patterns.mutable_index.mutable_index import MutableIndex


class MutableIndexTestCase(TestCase):
    dir_path = Path(os.path.realpath(__file__)).parent

    @unittest.skip("Not implemented")
    def test_find_mutable_index_assignment(self):
        self.assertEqual(
            [4],
            MutableIndex().value(Path(self.dir_path, "MutableIndexAssignment.java")),
            "Should match simple index assignment",
        )

    @unittest.skip("Not implemented")
    def test_find_mutable_index_prefix_increment(self):
        self.assertEqual(
            [4],
            MutableIndex().value(Path(self.dir_path, "MutableIndexPrefixIncrement.java")),
            "Should match index prefix increment (++i)",
        )

    @unittest.skip("Not implemented")
    def test_find_mutable_index_prefix_decrement(self):
        self.assertEqual(
            [4],
            MutableIndex().value(Path(self.dir_path, "MutableIndexPrefixDecrement.java")),
            "Should match index prefix decrement (--i)",
        )

    @unittest.skip("Not implemented")
    def test_find_mutable_index_postfix_increment(self):
        self.assertEqual(
            [4],
            MutableIndex().value(Path(self.dir_path, "MutableIndexPostfixIncrement.java")),
            "Should match index postfix increment (i++)",
        )

    @unittest.skip("Not implemented")
    def test_find_mutable_index_postfix_decrement(self):
        self.assertEqual(
            [4],
            MutableIndex().value(Path(self.dir_path, "MutableIndexPostfixDecrement.java")),
            "Should match index postfix decrement (i--)",
        )

    @unittest.skip("Not implemented")
    def test_find_mutable_index_addition_assignment(self):
        self.assertEqual(
            [4],
            MutableIndex().value(Path(self.dir_path, "MutableIndexAdditionAssignment.java")),
            "Should match index addition assignment (+=)",
        )

    @unittest.skip("Not implemented")
    def test_find_mutable_index_subtraction_assignment(self):
        self.assertEqual(
            [4],
            MutableIndex().value(Path(self.dir_path, "MutableIndexSubtractionAssignment.java")),
            "Should match index subtraction assignment (-=)",
        )

    @unittest.skip("Not implemented")
    def test_find_mutable_index_multiplication_assignment(self):
        self.assertEqual(
            [4],
            MutableIndex().value(Path(self.dir_path, "MutableIndexMultiplicationAssignment.java")),
            "Should match index multiplication assignment (*=)",
        )

    @unittest.skip("Not implemented")
    def test_find_mutable_index_division_assignment(self):
        self.assertEqual(
            [4],
            MutableIndex().value(Path(self.dir_path, "MutableIndexDivisionAssignment.java")),
            "Should match index division assignment (/=)",
        )

    @unittest.skip("Not implemented")
    def test_find_mutable_index_mod_assignment(self):
        self.assertEqual(
            [4],
            MutableIndex().value(Path(self.dir_path, "MutableIndexModAssignment.java")),
            "Should match index mod assignment (%=)",
        )

    @unittest.skip("Not implemented")
    def test_find_mutable_index_nested_for(self):
        self.assertEqual(
            [5],
            MutableIndex().value(Path(self.dir_path, "MutableIndexNestedFor.java")),
            "Should match mutable index in nested for",
        )

    @unittest.skip("Not implemented")
    def test_find_mutable_index_hidden_nested_for(self):
        self.assertEqual(
            [5],
            MutableIndex().value(Path(self.dir_path, "MutableIndexHiddenNestedFor.java")),
            "Should match mutable index hidden in nested for",
        )

    @unittest.skip("Not implemented")
    def test_find_mutable_index_no_block_for(self):
        self.assertEqual(
            [3],
            MutableIndex().value(Path(self.dir_path, "MutableIndexNoBlockFor.java")),
            "Should match mutable index in for with no block",
        )



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

from typing import List

from aibolit.ast_framework import AST, ASTNode, ASTNodeType


class ReturnNull:
    '''
    FInds all return statements that returns null directly or by ternary operator.
    NOTICE: nested ternary operators are not checked.
    '''

    def value(self, ast: AST) -> List[int]:
        lines: List[int] = []
        for return_statement in ast.get_proxy_nodes(ASTNodeType.RETURN_STATEMENT):
            if self._check_null_return_statement(return_statement):
                lines.append(return_statement.line)

        return lines

    def _check_null_return_statement(self, return_statement: ASTNode) -> bool:
        # return statement with no expression `return;` does not return null
        if return_statement.expression is None:
            return False

        if return_statement.expression.node_type == ASTNodeType.TERNARY_EXPRESSION:
            return self._check_null_expression(return_statement.expression.if_true) or \
                self._check_null_expression(return_statement.expression.if_false)

        return self._check_null_expression(return_statement.expression)

    def _check_null_expression(self, expression: ASTNode) -> bool:
        return expression.node_type == ASTNodeType.LITERAL and expression.value == 'null'



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

from pathlib import Path

from unittest import TestCase

from aibolit.metrics.fanout.FanOut import FanOut
from aibolit.ast_framework import AST
from aibolit.utils.ast_builder import build_ast


class FanOutTestSuite(TestCase):
    def test_count_types_in_fields(self):
        ast = FanOutTestSuite._build_ast('CountTypesInFields.java')
        fan_out = FanOut()
        self.assertEqual(fan_out.value(ast), 3)

    def test_type_used_several_times(self):
        ast = FanOutTestSuite._build_ast('TypeUsedSeveralTimes.java')
        fan_out = FanOut()
        self.assertEqual(fan_out.value(ast), 1)

    def test_self_usage(self):
        ast = FanOutTestSuite._build_ast('SelfUsage.java')
        fan_out = FanOut()
        self.assertEqual(fan_out.value(ast), 0)

    def test_class_referenced_from_package(self):
        ast = FanOutTestSuite._build_ast('ClassReferencedFromPackage.java')
        fan_out = FanOut()
        self.assertEqual(fan_out.value(ast), 2)

    def test_extending_type(self):
        ast = FanOutTestSuite._build_ast('ExtendingType.java')
        fan_out = FanOut()
        self.assertEqual(fan_out.value(ast), 1)

    def test_generic_type(self):
        ast = FanOutTestSuite._build_ast('GenericType.java')
        fan_out = FanOut()
        self.assertEqual(fan_out.value(ast), 1)

    def test_generic_and_packaged_types(self):
        ast = FanOutTestSuite._build_ast('GenericAndPackagedTypes.java')
        fan_out = FanOut()
        self.assertEqual(fan_out.value(ast), 2)

    @staticmethod
    def _build_ast(filename: str) -> AST:
        path = Path(__file__).absolute().parent / filename
        return AST.build_from_javalang(build_ast(str(path)))



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

import os
from unittest import TestCase
from aibolit.metrics.entropy.entropy import Entropy


class EntropyTest(TestCase):
    def test_it_works(self):
        entropy_metric = Entropy()
        entropy = entropy_metric.value(
            os.path.dirname(os.path.realpath(__file__)) + '/sample-1.java'
        )
        self.assertGreaterEqual(entropy, 0)



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT


class Loc:
    def __init__(self, path):
        self.path = path

    def value(self):
        i = -1
        with open(self.path, encoding='utf-8') as f:
            for i, l in enumerate(f):
                pass
        return i + 1



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

from collections import defaultdict
import sys
from typing import Dict, Callable, Any, TYPE_CHECKING

if TYPE_CHECKING:
    from aibolit.ast_framework import ASTNode, ASTNodeType  # noqa: F401


class _ComputedFieldsRegistry:
    def __init__(self) -> None:
        RegistryType = Dict["ASTNodeType", Dict[str, Callable[["ASTNode"], Any]]]
        self._registry: RegistryType = defaultdict(dict)

    def register(
        self,
        compute_field: Callable[["ASTNode"], Any],
        name: str,
        *node_types: "ASTNodeType",
    ) -> None:
        for node_type in node_types:
            computed_fields = self._registry[node_type]
            if name in computed_fields and \
               not _ComputedFieldsRegistry._is_in_interactive_shell():
                raise RuntimeError(
                    f"Registry already has computed field "
                    f"named '{name}' for node type {node_type}."
                )

            computed_fields[name] = compute_field

    def get_fields(
        self, node_type: "ASTNodeType"
    ) -> Dict[str, Callable[["ASTNode"], Any]]:
        return self._registry[node_type]

    def clear(self) -> None:
        self._registry = defaultdict(dict)

    @staticmethod
    def _is_in_interactive_shell() -> bool:
        """
        Taken from comments to this answer https://stackoverflow.com/a/6879085/2129920
        """
        return bool(getattr(sys, "ps1", sys.flags.interactive))


computed_fields_registry = _ComputedFieldsRegistry()



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

from pathlib import Path
from unittest import TestCase

from aibolit.patterns.classic_getter.classic_getter import ClassicGetter
from aibolit.ast_framework import AST
from aibolit.utils.ast_builder import build_ast


class SetterTestCase(TestCase):
    current_directory = Path(__file__).absolute().parent

    def test_no_getters(self):
        filepath = self.current_directory / "NoGetters.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = ClassicGetter()
        lines = pattern.value(ast)
        self.assertEqual(lines, [])

    def test_fake_getter(self):
        filepath = self.current_directory / "FakeGetter.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = ClassicGetter()
        lines = pattern.value(ast)
        self.assertEqual(lines, [])

    def test_long_fake_getter(self):
        filepath = self.current_directory / "LongFake.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = ClassicGetter()
        lines = pattern.value(ast)
        self.assertEqual(lines, [])

    def test_simple(self):
        filepath = self.current_directory / "SimpleGetter.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = ClassicGetter()
        lines = pattern.value(ast)
        self.assertEqual(lines, [8])



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

from typing import List

from aibolit.ast_framework import AST, ASTNodeType


class ArrayAsArgument:
    '''
    Finds all methods declarations, which accept arrays as parameters.
    '''

    def value(self, ast: AST) -> List[int]:
        lines: List[int] = []
        for method_declaration in ast.get_proxy_nodes(ASTNodeType.METHOD_DECLARATION,
                                                      ASTNodeType.CONSTRUCTOR_DECLARATION):
            if any(len(parameter.type.dimensions) > 0
                   for parameter in method_declaration.parameters):
                lines.append(method_declaration.line)
        return lines



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

from pathlib import Path
from unittest import TestCase

from aibolit.patterns.public_static_method.public_static_method import PublicStaticMethod
from aibolit.ast_framework import AST
from aibolit.utils.ast_builder import build_ast


class PublicStaticMethodPatternTestCase(TestCase):
    current_directory = Path(__file__).absolute().parent

    def test_find_non_final_attributes(self):
        filepath = self.current_directory / "PublicStaticMethod.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = PublicStaticMethod()
        lines = pattern.value(ast)
        self.assertEqual(len(lines), 1)



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT
import math
import os
from pathlib import Path
from unittest import TestCase, skip

import numpy as np
import pandas as pd

from aibolit.config import Config
from aibolit.model.stats import Stats
from aibolit.model.model import get_minimum, generate_fake_dataset


class TestStats(TestCase):

    def __init__(self, *args, **kwargs):
        super(TestStats, self).__init__(*args, **kwargs)
        self.cur_file_dir = Path(os.path.realpath(__file__)).parent
        self.config = Config.get_patterns_config()

    def test_get_minimum(self):
        minimum_arr = get_minimum([0, 0.23, 0.45], [0.34, 0.01, 0.37], [0.01, 0.50, 0.2])
        self.assertTrue(np.array_equal(minimum_arr[0], np.array([0, 0.01, 0.2])))
        self.assertTrue(np.array_equal(minimum_arr[1], np.array([0, 1, 2])))

    def test_get_array(self):
        lst = [np.array([0, 1, 2, 3, 4, 5], dtype=float),
               np.array([0, 1, 2, 3, 3, 3], dtype=float)]
        x = np.array(lst)
        mask = x > 0
        ncss = np.array([0.01, 0.02])
        res = Stats.change_matrix_by_value(x, mask, 2, ncss)
        self.assertTrue(
            np.array_equal(
                res,
                np.array([[0., 1., 2.01, 3., 4., 5.],
                          [0., 1., 2.02, 3., 3., 3.]])
            )
        )

    def test_split_dataset_by_pattern_value(self):
        x = [[0, 0, 0], [0, 0, 1], [1, 1, 2]]
        nulls, not_nulls = Stats.split_dataset_by_pattern_value(x, 2)
        self.assertTrue(np.array_equal(nulls[0], np.array([0, 0, 0])))
        self.assertTrue(np.array_equal(not_nulls[0], np.array([0, 0, 1])))
        self.assertTrue(np.array_equal(not_nulls[1], np.array([1, 1, 2])))

    def __load_mock_model(self):
        config = Config.get_patterns_config()
        patterns = [x['code'] for x in config['patterns']]

        class MockModel:

            def predict(self, input: np.array) -> np.array:
                results = []
                for row in input:
                    s = sum(row)
                    radian = math.radians(s)
                    results.append(math.sin(radian))
                return np.array(results)

        class PatternRankingModel:

            def __init__(self):
                self.features_conf = {
                    'features_order': patterns,
                    'patterns_only': patterns
                }
                self.model = MockModel()

        return PatternRankingModel()

    @skip("Skipping test due to np.bool_ assertion issue in CI")
    def test_stat_aibolit_pipeline(self):
        model = self.__load_mock_model()
        test_df = generate_fake_dataset()
        table = Stats.aibolit_stat(test_df, model)
        test_csv = Path(self.cur_file_dir, 'results_test.csv')
        results_df = pd.read_csv(test_csv, index_col=0)
        all_elements_compared: pd.DataFrame = table.eq(results_df)
        bool_eq_elems = np.ravel(all_elements_compared.values)
        are_equal_arrays = np.logical_and.reduce(bool_eq_elems, axis=0)
        self.assertTrue(are_equal_arrays)



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

from typing import Tuple

from networkx import DiGraph, disjoint_union  # type: ignore

from aibolit.ast_framework import AST, ASTNodeType


NODE_TYPES = [
    ASTNodeType.ASSIGNMENT,
    ASTNodeType.RETURN_STATEMENT
]


def build_cfg(tree: AST) -> DiGraph:
    '''Create Control Flow Graph'''
    g = DiGraph()
    g.add_node(0)
    for node in tree:
        if node.node_type not in NODE_TYPES:
            continue
        _g = _mk_cfg_graph(node.node_type)
        g = _compose_two_graphs(g, _g)
    return g


def _mk_cfg_graph(node: ASTNodeType) -> Tuple[DiGraph, int]:
    '''Takes in Javalang statement and returns corresponding CFG'''
    g = DiGraph()
    g.add_node(0)
    return g


def _compose_two_graphs(g1: DiGraph, g2: DiGraph) -> DiGraph:
    """Compose two graphs by creating edge between last of fist graph & fist of second.
       We assume that node in the each graph G has order from 0 to len(G)-1
    """
    g = disjoint_union(g1, g2)
    g.add_edge(len(g1) - 1, len(g1))
    return g



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT
from typing import TYPE_CHECKING, Dict

from cached_property import cached_property  # type: ignore
from deprecated import deprecated  # type: ignore
from networkx import DiGraph  # type: ignore

from aibolit.ast_framework import AST, ASTNodeType, ASTNode
from aibolit.utils.cfg_builder import build_cfg

if TYPE_CHECKING:
    from aibolit.ast_framework.java_class import JavaClass  # type: ignore


@deprecated(reason='This functionality must be transmitted to ASTNode')
class JavaClassMethod(AST):
    def __init__(self, tree: DiGraph, root: int, java_class: 'JavaClass'):
        super().__init__(tree, root)
        self._java_class = java_class

    def __str__(self):
        return f'Method {self.name}'

    @cached_property
    def name(self) -> str:
        try:
            root_node = ASTNode(self.tree, self.root)
            method_name_nodes = [
                child for child in root_node.children if child.node_type == ASTNodeType.STRING
            ]
            if not method_name_nodes:
                msg = "Provided AST does not has 'STRING' node type right under the root"
                raise ValueError(msg)
            return method_name_nodes[0].string
        except ValueError:
            raise ValueError("Provided AST does not has 'STRING' node type right under the root")

    @property
    def java_class(self) -> 'JavaClass':
        return self._java_class

    @cached_property
    def parameters(self) -> Dict[str, AST]:
        parameters: Dict[str, AST] = {}
        for parameter in self.get_subtrees(ASTNodeType.FORMAL_PARAMETER):
            parameter_node = ASTNode(self.tree, parameter.root)
            string_children = [
                child for child in parameter_node.children
                if child.node_type == ASTNodeType.STRING
            ]
            if string_children:
                parameter_name = string_children[0].string
                parameters[parameter_name] = parameter
        return parameters

    @cached_property
    def body(self) -> AST:
        return self.get_subtree(self.get_root())

    @cached_property
    def cfg(self) -> DiGraph:
        return build_cfg(self.body)



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

from pathlib import Path
from unittest import TestCase

from aibolit.patterns.nested_blocks.nested_blocks import NestedBlocks
from aibolit.ast_framework import AST, ASTNodeType
from aibolit.utils.ast_builder import build_ast


class NestedBlocksTestCase(TestCase):
    current_directory = Path(__file__).absolute().parent

    def test_single_for_loop(self):
        filepath = self.current_directory / "SingleFor.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = NestedBlocks(2, ASTNodeType.FOR_STATEMENT)
        lines = pattern.value(ast)
        self.assertEqual(lines, [18, 22])

    def test_nested_for_loops(self):
        filepath = self.current_directory / "NestedFor.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = NestedBlocks(2, ASTNodeType.FOR_STATEMENT)
        lines = pattern.value(ast)
        self.assertEqual(lines, [25])

    def test_for_loops_in_different_methods(self):
        filepath = self.current_directory / "DifferentMethods.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = NestedBlocks(2, ASTNodeType.FOR_STATEMENT)
        lines = pattern.value(ast)
        self.assertEqual(lines, [31])

    def test_for_loops_in_nested_class(self):
        filepath = self.current_directory / "NestedForInNestedClasses.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = NestedBlocks(2, ASTNodeType.FOR_STATEMENT)
        lines = pattern.value(ast)
        self.assertEqual(lines, [12])

    def test_for_loops_in_anonymous_class(self):
        filepath = self.current_directory / "ForInAnonymousFile.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = NestedBlocks(2, ASTNodeType.FOR_STATEMENT)
        lines = pattern.value(ast)
        self.assertEqual(lines, [22])

    def test_nested_no_nested_if(self):
        filepath = self.current_directory / "NestedNoIF.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = NestedBlocks(2, ASTNodeType.IF_STATEMENT)
        lines = pattern.value(ast)
        self.assertEqual(lines, [])

    def test_nested_if(self):
        filepath = self.current_directory / "NestedIF.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = NestedBlocks(2, ASTNodeType.IF_STATEMENT)
        lines = pattern.value(ast)
        self.assertEqual(lines, [24, 45])



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

import os
from unittest import TestCase
from pathlib import Path

from aibolit.patterns.empty_rethrow.empty_rethrow import EmptyRethrow
from aibolit.ast_framework import AST
from aibolit.utils.ast_builder import build_ast


class EmptyRethrowTestCase(TestCase):
    current_directory = Path(os.path.realpath(__file__)).parent

    def test_empty(self):
        filepath = str(Path(self.current_directory, "Empty.java"))
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = EmptyRethrow()
        self.assertEqual(pattern.value(ast), [])

    def test_anonymous(self):
        filepath = str(Path(self.current_directory, "Anonymous.java"))
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = EmptyRethrow()
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = EmptyRethrow()
        self.assertEqual(pattern.value(ast), [22, 28])

    def test_both_catches(self):
        filepath = str(Path(self.current_directory, "BothCatches.java"))
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = EmptyRethrow()
        self.assertEqual(pattern.value(ast), [9, 11])

    def test_instance_different_methods(self):
        filepath = str(Path(self.current_directory, "MultipleCatch.java"))
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = EmptyRethrow()
        self.assertEqual(pattern.value(ast), [9])

    def test_sequential_catch(self):
        filepath = str(Path(self.current_directory, "SequentialCatch.java"))
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = EmptyRethrow()
        self.assertEqual(pattern.value(ast), [9, 11])

    def test_sequential_catch_try(self):
        filepath = str(Path(self.current_directory, "SequentialCatchTry.java"))
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = EmptyRethrow()
        self.assertEqual(pattern.value(ast), [9, 11, 16, 18])

    def test_simple(self):
        filepath = str(Path(self.current_directory, "Simple.java"))
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = EmptyRethrow()
        self.assertEqual(pattern.value(ast), [9])

    def test_tricky_fake(self):
        filepath = str(Path(self.current_directory, "TrickyFake.java"))
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = EmptyRethrow()
        self.assertEqual(pattern.value(ast), [])

    def test_try_inside_catch(self):
        filepath = str(Path(self.current_directory, "TryInsideCatch.java"))
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = EmptyRethrow()
        self.assertEqual(pattern.value(ast), [13])

    def test_try_inside_finally(self):
        filepath = str(Path(self.current_directory, "TryInsideFinally.java"))
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = EmptyRethrow()
        self.assertEqual(pattern.value(ast), [9, 15])

    def test_try_inside_try(self):
        filepath = str(Path(self.current_directory, "TryInsideTry.java"))
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = EmptyRethrow()
        self.assertEqual(pattern.value(ast), [13, 20])

    def test_catch_with_functions(self):
        filepath = str(Path(self.current_directory, "CatchWithFunctions.java"))
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = EmptyRethrow()
        self.assertEqual(pattern.value(ast), [16])

    def test_catch_with_if(self):
        filepath = str(Path(self.current_directory, "CatchWithIf.java"))
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = EmptyRethrow()
        self.assertEqual(pattern.value(ast), [17])

    def test_chained_exceptions1(self):
        filepath = str(Path(self.current_directory, "DataBaseNavigator.java"))
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = EmptyRethrow()
        self.assertEqual(pattern.value(ast), [])

    def test_chained_exceptions2(self):
        filepath = str(Path(self.current_directory, "ConfigImportWizardPageDbvis.java"))
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = EmptyRethrow()
        self.assertEqual(pattern.value(ast), [])

    def test_chained_exceptions3(self):
        filepath = str(Path(self.current_directory, "DatabaseNavigatorSourceContainer.java"))
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = EmptyRethrow()
        self.assertEqual(pattern.value(ast), [])

    def test_try_without_catch(self):
        filepath = str(Path(self.current_directory, "ConcurrentDiskUtil.java"))
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = EmptyRethrow()
        self.assertEqual(pattern.value(ast), [])

    def test_throw_with_cast(self):
        filepath = str(Path(self.current_directory, "CommonRdbmsReader.java"))
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = EmptyRethrow()
        self.assertEqual(pattern.value(ast), [])



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

from typing import Set, NamedTuple

from aibolit.ast_framework import AST, ASTNode, ASTNodeType


class _MethodInvocationParams(NamedTuple):
    isLocal: bool
    name: str


class RFC:
    """
    The Response For a Class (RFC) is an object-oriented metric
    that shows the interaction of the class methods with other methods.
    This implementation accept arbitrary AST and
    return sum of RFC for all class declaration in it.
    To calculate RFC of a class we count number of public methods and
    number of distinct methods invocation in those methods.
    """

    def value(self, ast: AST) -> int:
        rfc = 0
        for class_declaration in ast.get_proxy_nodes(ASTNodeType.CLASS_DECLARATION):
            rfc += self._calculate_class_RFC(ast.get_subtree(class_declaration))

        return rfc

    def _calculate_class_RFC(self, java_class: AST) -> int:
        class_declaration = java_class.get_root()
        assert class_declaration.node_type == ASTNodeType.CLASS_DECLARATION

        rfc = 0
        invoked_methods: Set[_MethodInvocationParams] = set()
        local_methods_names: Set[str] = set()
        for method_ast in java_class.get_subtrees(ASTNodeType.METHOD_DECLARATION):
            method_declaration = method_ast.get_root()
            local_methods_names.add(method_declaration.name)
            if "public" in method_declaration.modifiers:
                rfc += 1
                invoked_methods |= self._get_all_method_invocation_params(method_ast)

        # filter out inherited methods
        # consider local methods with name not found
        # among methods names of current class as inherited
        invoked_methods = {
            invoked_method
            for invoked_method in invoked_methods
            if not invoked_method.isLocal or invoked_method.name in local_methods_names
        }

        rfc += len(invoked_methods)
        return rfc

    def _get_all_method_invocation_params(self, ast: AST) -> Set[_MethodInvocationParams]:
        return {
            self._create_method_invocation_params(method_invocation)
            for method_invocation in ast.get_proxy_nodes(ASTNodeType.METHOD_INVOCATION)
        }

    def _create_method_invocation_params(self,
                                         method_invocation: ASTNode) -> _MethodInvocationParams:
        assert method_invocation.node_type == ASTNodeType.METHOD_INVOCATION
        return _MethodInvocationParams(
            isLocal=method_invocation.qualifier is None, name=method_invocation.member
        )



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

"""Type declarations module for aibolit.

This module provides type definitions used throughout the aibolit codebase.
"""

LineNumber = int



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

from pathlib import Path
from unittest import TestCase

from aibolit.patterns.null_check.null_check import NullCheck
from aibolit.ast_framework import AST
from aibolit.utils.ast_builder import build_ast


class NullCheckTestCase(TestCase):
    current_directory = Path(__file__).absolute().parent

    def test_null_check(self):
        filepath = self.current_directory / "1.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = NullCheck()
        lines = pattern.value(ast)
        self.assertEqual(lines, [7])

    def test_null_check_in_constructor(self):
        filepath = self.current_directory / "2.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = NullCheck()
        lines = pattern.value(ast)
        self.assertEqual(lines, [])

    def test_null_check_comparison_result_assignment(self):
        filepath = self.current_directory / "3.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = NullCheck()
        lines = pattern.value(ast)
        self.assertEqual(lines, [7])

    def test_null_check_ternary(self):
        filepath = self.current_directory / "4.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = NullCheck()
        lines = pattern.value(ast)
        self.assertEqual(lines, [7])

    def test_null_check_not_equal_comparison(self):
        filepath = self.current_directory / "5.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = NullCheck()
        lines = pattern.value(ast)
        self.assertEqual(lines, [7])



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

import os
import unittest
from pathlib import Path
from unittest import TestCase

from aibolit.patterns.non_final_argument.non_final_argument import NonFinalArgument


@unittest.skip("Not implemented")
class NonFinalArgumentTestCase(TestCase):
    dir_path = Path(os.path.realpath(__file__)).parent

    def test_find_non_final_argument_in_ctor(self):
        self.assertEqual(
            NonFinalArgument().value(Path(self.dir_path, "NonFinalArgumentCtor.java")), [7]
        )

    def test_find_non_final_argument_in_method(self):
        self.assertEqual(
            NonFinalArgument().value(Path(self.dir_path, "NonFinalArgumentMethod.java")), [11]
        )



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

from pathlib import Path
from unittest import TestCase

from aibolit.patterns.multiple_while.multiple_while import MultipleWhile
from aibolit.ast_framework import AST
from aibolit.utils.ast_builder import build_ast


class MultipleWhilePatternTestCase(TestCase):
    current_directory = Path(__file__).absolute().parent

    def test_simple(self):
        filepath = self.current_directory / "Simple.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = MultipleWhile()
        lines = pattern.value(ast)
        self.assertEqual(lines, [5])

    def test_one_while(self):
        filepath = self.current_directory / "OneWhile.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = MultipleWhile()
        lines = pattern.value(ast)
        self.assertEqual(lines, [])

    def test_if_while(self):
        filepath = self.current_directory / "IfWhile.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = MultipleWhile()
        lines = pattern.value(ast)
        self.assertEqual(lines, [5])

    def test_multiple_while(self):
        filepath = self.current_directory / "MultipleWhile.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = MultipleWhile()
        lines = pattern.value(ast)
        self.assertEqual(lines, [])



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

import os
from unittest import TestCase
from pathlib import Path

from aibolit.patterns.array_as_argument.array_as_argument import ArrayAsArgument
from aibolit.ast_framework import AST
from aibolit.utils.ast_builder import build_ast


class ArrayAsArgumentTestCase(TestCase):
    dir_path = Path(os.path.realpath(__file__)).parent
    pattern = ArrayAsArgument()

    def test_NoArgument(self):
        file = Path(self.dir_path, "NoArgument.java")
        ast = AST.build_from_javalang(build_ast(file))
        self.assertEqual([], self.pattern.value(ast), "Should not match no argument method")

    def test_PrimitiveAsArgument(self):
        file = Path(self.dir_path, "PrimitiveAsArgument.java")
        ast = AST.build_from_javalang(build_ast(file))
        self.assertEqual(
            [], self.pattern.value(ast), "Should not match method with primitive as argument"
        )

    def test_ArrayAsArgument(self):
        file = Path(self.dir_path, "ArrayAsArgument.java")
        ast = AST.build_from_javalang(build_ast(file))
        self.assertEqual([5], self.pattern.value(ast), "Should match method with array as argument")

    def test_ObjectAsArgument(self):
        file = Path(self.dir_path, "ObjectAsArgument.java")
        ast = AST.build_from_javalang(build_ast(file))
        self.assertEqual(
            [], self.pattern.value(ast), "Should not match method with object as argument"
        )

    def test_PrimitiveAndArrayAsArgument(self):
        file = Path(self.dir_path, "PrimitiveAndArrayAsArgument.java")
        ast = AST.build_from_javalang(build_ast(file))
        self.assertEqual([5], self.pattern.value(ast), "Should match method with array as argument")

    def test_GenericArrayAsArgument(self):
        file = Path(self.dir_path, "GenericArrayAsArgument.java")
        ast = AST.build_from_javalang(build_ast(file))
        self.assertEqual(
            [5], self.pattern.value(ast), "Should match method with generic array as argument"
        )

    def test_ConstructorWithArrayAsArgument(self):
        file = Path(self.dir_path, "ConstructorWithArrayAsArgument.java")
        ast = AST.build_from_javalang(build_ast(file))
        self.assertEqual(
            [5], self.pattern.value(ast), "Should match constructor with array as argument"
        )



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

from unittest import TestCase
from pathlib import Path

from aibolit.patterns.method_chaining.method_chaining import MethodChainFind
from aibolit.ast_framework import AST
from aibolit.utils.ast_builder import build_ast


class MethodChainTestCase(TestCase):
    current_directory = Path(__file__).absolute().parent

    def test_method_chain(self):
        filepath = self.current_directory / "MethodChain.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = MethodChainFind()
        lines = pattern.value(ast)
        self.assertEqual(lines, [24])

    def test_empty_method_chain(self):
        filepath = self.current_directory / "EmptyMethodChain.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = MethodChainFind()
        lines = pattern.value(ast)
        self.assertEqual(lines, [24])

    def test_chain_with_new_object(self):
        filepath = self.current_directory / "MethodChainNewObjectMethods.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = MethodChainFind()
        lines = pattern.value(ast)
        self.assertEqual(lines, [26, 37])

    def test_method_chain_in_different_methods(self):
        filepath = self.current_directory / "MethodChainInDifferentMethods.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = MethodChainFind()
        lines = pattern.value(ast)
        self.assertEqual(lines, [25, 37])

    def test_chain_in_nested_class(self):
        filepath = self.current_directory / "MethodChainNestedClass.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = MethodChainFind()
        lines = pattern.value(ast)
        self.assertEqual(lines, [22])

    def test_chain_in_anonymous_class(self):
        filepath = self.current_directory / "MethodChainAnonymousClass.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = MethodChainFind()
        lines = pattern.value(ast)
        self.assertEqual(lines, [32])

    def test_chain_in_anonymous_class_empty(self):
        filepath = self.current_directory / "MethodChainAnonymousClassEmpty.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = MethodChainFind()
        lines = pattern.value(ast)
        self.assertEqual(lines, [])

    def test_several_chains(self):
        filepath = self.current_directory / "MethodChainSeveral.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = MethodChainFind()
        lines = pattern.value(ast)
        self.assertEqual(lines, [16, 37, 51])

    def test_chain_without_object_creating(self):
        filepath = self.current_directory / "WithoutObjectCreating.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = MethodChainFind()
        lines = pattern.value(ast)
        self.assertEqual(lines, [17])

    def test_nested_chain_with_this(self):
        filepath = self.current_directory / "NestedChainWIthThis.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = MethodChainFind()
        lines = pattern.value(ast)
        self.assertEqual(lines, [17, 18])

    def test_nested_chain_with_simple_method_invocation(self):
        filepath = self.current_directory / "NestedChainWithSimpleMethodInvocation.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = MethodChainFind()
        lines = pattern.value(ast)
        self.assertEqual(lines, [18, 19])

    def test_nested_chain_complicated_structure(self):
        """
        Several nested structures are checked: nested method chaining
        with nested anonymous classes
        """
        filepath = self.current_directory / "HolyMolyNestedChain.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = MethodChainFind()
        lines = pattern.value(ast)
        self.assertEqual(lines, [63, 70, 80])

    def test_smallest_chain(self):
        filepath = self.current_directory / "SmallestChain.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = MethodChainFind()
        lines = pattern.value(ast)
        self.assertEqual(lines, [33, 85, 86])

    def test_fake_chain(self):
        filepath = self.current_directory / "FakeChain.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = MethodChainFind()
        lines = pattern.value(ast)
        self.assertEqual(lines, [])

    def test_many_chains(self):
        filepath = self.current_directory / "MachineLearningGetResultsIT.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = MethodChainFind()
        lines = pattern.value(ast)
        self.assertGreater(len(lines), 300)



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT
# import os
from unittest import TestCase
# from aibolit.config import Config


class TestConfig(TestCase):
    pass



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

import os
import unittest
from pathlib import Path
from unittest import TestCase
from aibolit.patterns.loop_outsider.loop_outsider import LoopOutsider


@unittest.skip("Not implemented")
class LoopOutsiderTestCase(TestCase):
    """
    @todo #138:30min Continue to implement loop_outsider
     Implement loop_outsider tests regarding for and do-while loops. Then implement
     loop_outsider pattern itself.
    """

    dir_path = Path(os.path.realpath(__file__)).parent

    def test_find_loop_outsider_assignment_in_while(self):
        self.assertEqual(
            LoopOutsider().value(Path(self.dir_path, "LoopOutsiderAssignmentInWhile.java")),
            [5],
            "Could not find loop outsider assignment in while loop",
        )

    def test_find_loop_outsider_prefix_increment_in_while(self):
        self.assertEqual(
            LoopOutsider().value(Path(self.dir_path, "LoopOutsiderPrefixIncrementInWhile.java")),
            [5],
            "Could not find loop outsider with prefix increment in while loop",
        )

    def test_find_loop_outsider_postfix_increment_in_while(self):
        self.assertEqual(
            LoopOutsider().value(Path(self.dir_path, "LoopOutsiderPostfixIncrementInWhile.java")),
            [5],
            "Could not find loop outsider with postfix increment in while loop",
        )

    def test_find_loop_outsider_prefix_decrement_in_while(self):
        self.assertEqual(
            LoopOutsider().value(Path(self.dir_path, "LoopOutsiderPrefixDecrementInWhile.java")),
            [5],
            "Could not find loop outsider with prefix decrement in while loop",
        )

    def test_find_loop_outsider_postfix_decrement_in_while(self):
        self.assertEqual(
            LoopOutsider().value(Path(self.dir_path, "LoopOutsiderPostfixDecrementInWhile.java")),
            [5],
            "Could not find loop outsider with postfix decrement in while loop",
        )

    def test_find_loop_outsider_add_and_in_while(self):
        self.assertEqual(
            LoopOutsider().value(Path(self.dir_path, "LoopOutsiderAddAndInWhile.java")),
            [5],
            "Could not find loop outsider with add AND in while loop",
        )

    def test_find_loop_outsider_subtract_and_in_while(self):
        self.assertEqual(
            LoopOutsider().value(Path(self.dir_path, "LoopOutsiderSubtractAndInWhile.java")),
            [5],
            "Could not find loop outsider with subtract AND in while loop",
        )

    def test_find_loop_outsider_multiply_and_in_while(self):
        self.assertEqual(
            LoopOutsider().value(Path(self.dir_path, "LoopOutsiderMultiplyAndInWhile.java")),
            [5],
            "Could not find loop outsider with multiply AND in while loop",
        )

    def test_find_loop_outsider_divide_and_in_while(self):
        self.assertEqual(
            LoopOutsider().value(Path(self.dir_path, "LoopOutsiderDivideAndInWhile.java")),
            [5],
            "Could not find loop outsider with divide AND in while loop",
        )

    def test_find_loop_outsider_modulus_and_in_while(self):
        self.assertEqual(
            LoopOutsider().value(Path(self.dir_path, "LoopOutsiderModulusAndInWhile.java")),
            [5],
            "Could not find loop outsider with modulus AND in while loop",
        )

    def test_find_loop_outsider_left_shift_and_in_while(self):
        self.assertEqual(
            LoopOutsider().value(Path(self.dir_path, "LoopOutsiderLeftShiftAndInWhile.java")),
            [5],
            "Could not find loop outsider with left shift AND in while loop",
        )

    def test_find_loop_outsider_right_shift_and_in_while(self):
        self.assertEqual(
            LoopOutsider().value(Path(self.dir_path, "LoopOutsiderRightShiftAndInWhile.java")),
            [5],
            "Could not find loop outsider with right shift AND in while loop",
        )

    def test_find_loop_outsider_bitwise_and_in_while(self):
        self.assertEqual(
            LoopOutsider().value(Path(self.dir_path, "LoopOutsiderBitwiseAndInWhile.java")),
            [5],
            "Could not find loop outsider with bitwise AND in while loop",
        )

    def test_find_loop_outsider_bitwise_inclusive_or_in_while(self):
        self.assertEqual(
            LoopOutsider().value(Path(self.dir_path, "LoopOutsiderBitwiseInclusiveOrInWhile.java")),
            [5],
            "Could not find loop outsider with bitwise inclusive OR in while loop",
        )

    def test_find_loop_outsider_bitwise_exclusive_or_in_while(self):
        self.assertEqual(
            LoopOutsider().value(Path(self.dir_path, "LoopOutsiderBitwiseExclusiveOrInWhile.java")),
            [5],
            "Could not find loop outsider with bitwise exclusive OR in while loop",
        )



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

from pathlib import Path
from unittest import TestCase

from aibolit.patterns.joined_validation.joined_validation import JoinedValidation
from aibolit.ast_framework import AST
from aibolit.utils.ast_builder import build_ast


class JoinedValidationTestCase(TestCase):
    current_directory = Path(__file__).absolute().parent

    def test_canFindSimpleJoinedValidation(self):
        filepath = self.current_directory / "SimpleJoinedValidation.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = JoinedValidation()
        lines = pattern.value(ast)
        self.assertEqual([6], lines, "Could not find simple joined validation")

    def test_canFindJoinedValidationAndOr(self):
        filepath = self.current_directory / "JoinedValidationAndOr.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = JoinedValidation()
        lines = pattern.value(ast)
        self.assertEqual([6], lines, "Could not find joined validation in AndOr condition")

    def test_canFindJoinedValidationOrAnd(self):
        filepath = self.current_directory / "JoinedValidationOrAnd.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = JoinedValidation()
        lines = pattern.value(ast)
        self.assertEqual([6], lines, "Could not find joined validation in OrAnd condition")

    def test_canFindJoinedValidationOrOr(self):
        filepath = self.current_directory / "JoinedValidationOrOr.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = JoinedValidation()
        lines = pattern.value(ast)
        self.assertEqual([6], lines, "Could not find joined validation in OrOr condition")

    def test_canFindJoinedValidationOrFunctionCall(self):
        filepath = self.current_directory / "JoinedValidationOrFunctionCall.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = JoinedValidation()
        lines = pattern.value(ast)
        self.assertEqual([11], lines, "Could not find joined validation in function call")

    def test_canFindJoinedValidationOrFieldAccess(self):
        filepath = self.current_directory / "JoinedValidationOrFieldAccess.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = JoinedValidation()
        lines = pattern.value(ast)
        self.assertEqual([9], lines, "Could not find joined validation in field access")

    def test_canFindNoBracketsJoinedValidation(self):
        filepath = self.current_directory / "NoBracketsJoinedValidation.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = JoinedValidation()
        lines = pattern.value(ast)
        self.assertEqual([6], lines, "Could not find joined validation when using no brackets")

    def test_canSkipEmptyJoinedValidation(self):
        filepath = self.current_directory / "EmptyJoinedValidation.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = JoinedValidation()
        lines = pattern.value(ast)
        self.assertEqual([], lines, "Could not skip empty joined validation")

    def test_canSkipNoJoinedValidation(self):
        filepath = self.current_directory / "NoJoinedValidation.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = JoinedValidation()
        lines = pattern.value(ast)
        self.assertEqual([], lines, "Could not skip when there is no joined validation")



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT
from typing import List, Any

from aibolit.ast_framework import ASTNodeType, AST


class SendNull:

    def __init__(self):
        pass

    def __is_null(self, val: Any) -> bool:
        if not hasattr(val, 'value'):
            return False
        if not isinstance(val.value, str):
            return False
        if val.value != 'null':
            return False
        return True

    def value(self, ast: AST) -> List[int]:

        lines = set()
        invocatios_types = [
            ASTNodeType.METHOD_INVOCATION,
            ASTNodeType.EXPLICIT_CONSTRUCTOR_INVOCATION,
            ASTNodeType.CLASS_CREATOR
        ]
        for node in ast.get_proxy_nodes(*invocatios_types):
            for argument in node.arguments:
                if (argument.node_type == ASTNodeType.LITERAL) and (argument.value == "null"):
                    lines.add(argument.line)

        for node in ast.get_proxy_nodes(ASTNodeType.TERNARY_EXPRESSION):
            if self.__is_null(node.if_false) or self.__is_null(node.if_true):
                lines.add(node.line)

        lst = sorted(lines)
        return lst



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

from unittest import TestCase
from pathlib import Path

from aibolit.metrics.RFC.rfc import RFC
from aibolit.ast_framework import AST
from aibolit.utils.ast_builder import build_ast


class RFCTestSuite(TestCase):
    def test_empty_public_methods(self):
        ast = RFCTestSuite._get_ast("EmptyPublicMethods.java")
        rfc = RFC()
        self.assertEqual(rfc.value(ast), 2)

    def test_empty_public_and_private_methods(self):
        ast = RFCTestSuite._get_ast("EmptyPublicAndPrivateMethods.java")
        rfc = RFC()
        self.assertEqual(rfc.value(ast), 2)

    def test_public_methods_invoke_public_methods(self):
        ast = RFCTestSuite._get_ast("PublicMethodsInvokePublicMethods.java")
        rfc = RFC()
        self.assertEqual(rfc.value(ast), 3)

    def test_public_methods_invoke_private_methods(self):
        ast = RFCTestSuite._get_ast("PublicMethodsInvokePrivateMethods.java")
        rfc = RFC()
        self.assertEqual(rfc.value(ast), 3)

    def test_public_methods_invoke_outer_methods(self):
        ast = RFCTestSuite._get_ast("PublicMethodsInvokeOuterMethods.java")
        rfc = RFC()
        self.assertEqual(rfc.value(ast), 3)

    def test_several_invocation_of_same_method(self):
        ast = RFCTestSuite._get_ast("SeveralInvocationOfSameMethod.java")
        rfc = RFC()
        self.assertEqual(rfc.value(ast), 4)

    def test_invocation_of_local_and_outer_methods_with_same_name(self):
        ast = RFCTestSuite._get_ast("InvocationOfLocalAndOuterMethodsWithSameName.java")
        rfc = RFC()
        self.assertEqual(rfc.value(ast), 4)

    def test_several_classes(self):
        ast = RFCTestSuite._get_ast("SeveralClasses.java")
        rfc = RFC()
        self.assertEqual(rfc.value(ast), 2)

    def test_inherited_methods(self):
        ast = RFCTestSuite._get_ast("InheritedMethods.java")
        rfc = RFC()
        self.assertEqual(rfc.value(ast), 2)

    def test_overwritten_inherited_methods(self):
        ast = RFCTestSuite._get_ast("OverwrittenInheritedMethods.java")
        rfc = RFC()
        self.assertEqual(rfc.value(ast), 2)

    @staticmethod
    def _get_ast(filename: str) -> AST:
        path = Path(__file__).absolute().parent / filename
        return AST.build_from_javalang(build_ast(str(path)))



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

from typing import Any, List, Iterator, Optional

from networkx import DiGraph, dfs_preorder_nodes  # type: ignore
from cached_property import cached_property  # type: ignore

from aibolit.ast_framework._auxiliary_data import (
    common_attributes,
    attributes_by_node_type,
    ASTNodeReference,
)
from aibolit.ast_framework import ASTNodeType
from aibolit.ast_framework.computed_fields_registry import computed_fields_registry


class ASTNode:
    def __init__(self, graph: DiGraph, node_index: int):
        self._graph = graph
        self._node_index = node_index

    @property
    def children(self) -> Iterator["ASTNode"]:
        for child_index in self._graph.succ[self._node_index]:
            yield ASTNode(self._graph, child_index)

    @property
    def parent(self) -> Optional["ASTNode"]:
        try:
            parent_index = next(self._graph.predecessors(self._node_index))
            return ASTNode(self._graph, parent_index)
        except StopIteration:
            return None

    @property
    def node_index(self) -> int:
        return self._node_index

    @cached_property
    def line(self) -> int:
        line = self._get_line(self._node_index)
        if line is not None:
            return line

        children_lines: List[int] = [
            self._get_line(child_index)  # type: ignore
            for child_index in dfs_preorder_nodes(self._graph, self._node_index)
            if self._get_line(child_index) is not None
        ]

        # try to find source code line information from nodes reachable from self
        if children_lines:
            return min(children_lines)

        # try to  find source code line information from parents up to the root
        parent_index = self._get_parent(self._node_index)
        while line is None and parent_index is not None:
            line = self._get_line(parent_index)
            parent_index = self._get_parent(parent_index)

        if line is not None:
            return line

        raise RuntimeError(
            f"Failed to retrieve source code line information for {repr(self)} node. "
            "All nodes in a path from root to it and all nodes, reachable from it, "
            "does not have any source code line information either."
        )

    def __getattr__(self, attribute_name: str):
        node_type = self._get_type(self._node_index)
        javalang_fields = attributes_by_node_type[node_type]
        computed_fields = computed_fields_registry.get_fields(node_type)
        if attribute_name not in common_attributes and \
           attribute_name not in javalang_fields and \
           attribute_name not in computed_fields:
            raise AttributeError(
                "Failed to retrieve property. "
                f"'{node_type}' node does not have '{attribute_name}' attribute."
            )

        if attribute_name in computed_fields:
            attribute = computed_fields[attribute_name](self)
        else:
            attribute = self._graph.nodes[self._node_index][attribute_name]

        # common_attributes and javalang_fields may contain ASTNodeReference
        # which needs to be replaces with actual ASTNode for convince API
        # computed_fields are using javalang_fields and there is no ASTNodeReference
        if attribute_name not in computed_fields:
            if isinstance(attribute, ASTNodeReference):
                attribute = self._create_node_from_reference(attribute)
            elif isinstance(attribute, list):
                attribute = self._replace_references_with_nodes(attribute)
        return attribute

    def __dir__(self) -> List[str]:
        node_type = self._get_type(self._node_index)
        return ASTNode._public_fixed_interface + \
            list(common_attributes) + \
            list(attributes_by_node_type[node_type]) + \
            list(computed_fields_registry.get_fields(node_type).keys())

    def __str__(self) -> str:
        text_representation = f"node index: {self._node_index}"
        node_type = self._get_type(self._node_index)
        for attribute_name in sorted(
            common_attributes | attributes_by_node_type[node_type]
        ):
            attribute_value = self.__getattr__(attribute_name)

            if isinstance(attribute_value, ASTNode):
                attribute_representation = repr(attribute_value)
            elif isinstance(attribute_value, str) and "\n" in attribute_value:
                attribute_representation = "\n\t" + attribute_value.replace(
                    "\n", "\n\t"
                )
            else:
                attribute_representation = str(attribute_value)

            text_representation += f"\n{attribute_name}: {attribute_representation}"

        return text_representation

    def __repr__(self) -> str:
        node_type = self._get_type(self._node_index)
        return f"<ASTNode node_type: {node_type}, node_index: {self._node_index}>"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ASTNode):
            msg = "ASTNode support comparission only with themselves, "
            msg += f"but {type(other)} was provided."
            raise NotImplementedError(msg)
        return self._graph == other._graph and self._node_index == other._node_index

    def __hash__(self):
        return hash(self._node_index)

    def _replace_references_with_nodes(
        self, list_with_references: List[Any]
    ) -> List[Any]:
        list_with_nodes: List[Any] = []
        for item in list_with_references:
            if isinstance(item, ASTNodeReference):
                list_with_nodes.append(self._create_node_from_reference(item))
            elif isinstance(item, list):
                list_with_nodes.append(self._replace_references_with_nodes(item))
            elif isinstance(item, (int, str)) or item is None:
                list_with_nodes.append(item)
            else:
                raise RuntimeError(
                    "Failed parsing attribute.\n"
                    f"An {item} with {type(item)} was found.\n"
                    "Expected: int, str, ASNodeReference of list of them."
                )

        return list_with_nodes

    def _create_node_from_reference(self, reference: ASTNodeReference) -> "ASTNode":
        return ASTNode(self._graph, reference.node_index)

    def _get_type(self, node_index: int) -> ASTNodeType:
        return self._graph.nodes[node_index]["node_type"]

    def _get_line(self, node_index: int) -> Optional[int]:
        return self._graph.nodes[node_index]["line"]

    def _get_parent(self, node_index: int) -> Optional[int]:
        # there is maximum one parent in a tree
        return next(self._graph.predecessors(node_index), None)

    # names of methods and properties, which is not generated dynamically
    _public_fixed_interface = ["children", "node_index", "line"]



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT
from typing import List, Union

from aibolit.ast_framework import ASTNodeType, AST
from aibolit.ast_framework.ast_node import ASTNode


class ManyPrimaryCtors:
    '''
    If there is more than one primary
    constructors in a class, it is
    considered a pattern
    '''
    def value(self, ast: AST) -> List[int]:
        lines: List[int] = []
        for class_declaration in ast.get_proxy_nodes(ASTNodeType.CLASS_DECLARATION):
            primary_lines = self.__find_primary(ast, class_declaration.body)
            if len(primary_lines) > 1:
                lines.extend(primary_lines)
        return lines

    def __find_primary(self, ast: AST, class_body: List[ASTNode]) -> List[int]:
        lines: List[int] = []
        for node in class_body:
            if self.__check_primary(ast, node):
                lines.append(node.line)
        return lines

    def __check_primary(self, ast: AST, node: Union[ASTNode, List[ASTNode]]) -> bool:
        if isinstance(node, ASTNode) and node.node_type == ASTNodeType.CONSTRUCTOR_DECLARATION:
            for assignment in ast.get_subtree(node).get_proxy_nodes(ASTNodeType.ASSIGNMENT):
                if assignment.expressionl.node_type == ASTNodeType.THIS:
                    return True
        return False



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

"""
Aibolit: Static Analyzer with Machine Learning in Mind
"""

__version__ = '1.3.0'
__author__ = 'Yegor Bugayenko'
__licence__ = 'MIT'



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT


from setuptools import setup, find_packages
import aibolit

setup(
    name='aibolit',
    version=aibolit.__version__,
    description=aibolit.__doc__.strip(),
    long_description='DevTool recommending how to improve the maintenance quality of '
                     'your Java classes',
    url='https://github.com/yegor256/aibolit',
    download_url='https://github.com/yegor256/aibolit',
    author=aibolit.__author__,
    author_email='yegor256@gmail.com',
    license=aibolit.__licence__,
    packages=find_packages(),
    entry_points={
        'console_scripts': [
            'aibolit = aibolit.__main__:main'
        ],
    },
    extras_require={},
    install_requires=open('requirements.txt', 'r', encoding='utf-8').readlines(),
    tests_require=open('requirements.txt', 'r', encoding='utf-8').readlines(),
    classifiers=[
        'Programming Language :: Python',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Topic :: Software Development',
        'Topic :: Utilities'
    ],
    include_package_data=True,
    package_data={
        'aibolit': [
            'binary_files/halstead.jar',
            'binary_files/model.dat']
    },
)



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

from pathlib import Path
from unittest import TestCase

from aibolit.metrics.NumberMethods.NumberMethods import NumberMethods
from aibolit.ast_framework import AST
from aibolit.utils.ast_builder import build_ast


class TestCognitive(TestCase):
    current_directory = Path(__file__).absolute().parent

    def test_one(self):
        filepath = self.current_directory / 'one.java'
        ast = AST.build_from_javalang(build_ast(filepath))
        metric = NumberMethods()
        value = metric.value(ast)
        self.assertEqual(value, 1)

    def test_simple(self):
        filepath = self.current_directory / 'simple.java'
        ast = AST.build_from_javalang(build_ast(filepath))
        metric = NumberMethods()
        value = metric.value(ast)
        self.assertEqual(value, 2)

    def test_nested(self):
        filepath = self.current_directory / 'nested.java'
        ast = AST.build_from_javalang(build_ast(filepath))
        metric = NumberMethods()
        value = metric.value(ast)
        self.assertEqual(value, 2)

    def test_several(self):
        filepath = self.current_directory / 'several.java'
        ast = AST.build_from_javalang(build_ast(filepath))
        metric = NumberMethods()
        value = metric.value(ast)
        self.assertEqual(value, 4)



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT
from typing import List

from aibolit.ast_framework import ASTNodeType, AST


class MultipleTry:
    '''
    Check if a method contains more than one Try Statement
    '''
    def value(self, ast: AST) -> List[int]:
        total_code_lines: List[int] = []
        for method_declaration in ast.get_proxy_nodes(ASTNodeType.METHOD_DECLARATION):
            try_nodes = list(ast.get_subtree(method_declaration).get_proxy_nodes(
                ASTNodeType.TRY_STATEMENT))
            if len(try_nodes) > 1:
                total_code_lines.append(method_declaration.line)
        return total_code_lines



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT
from typing import List
from aibolit.types_decl import LineNumber


class LoopOutsider:
    """
    Pattern which matches loop outsiders: when we modify a variable which is declared outside of the
    scope of the loop.
    """

    def __init__(self):
        pass

    def value(self, filename) -> List[LineNumber]:
        """
        Returns the line number of loop outsiders found in file.
        """
        return []



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

from pathlib import Path
from unittest import TestCase

from aibolit.patterns.instanceof.instance_of import InstanceOf
from aibolit.ast_framework import AST
from aibolit.utils.ast_builder import build_ast


class InstanceOfTestCase(TestCase):
    current_directory = Path(__file__).absolute().parent

    def test_empty(self):
        filepath = self.current_directory / "Empty.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = InstanceOf()
        lines = pattern.value(ast)
        self.assertEqual(len(lines), 0)

    def test_instance_of(self):
        filepath = self.current_directory / "InstanceOfSample.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = InstanceOf()
        lines = pattern.value(ast)
        self.assertEqual(len(lines), 1)

    def test_instance(self):
        filepath = self.current_directory / "InstanceSample.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = InstanceOf()
        lines = pattern.value(ast)
        self.assertEqual(len(lines), 1)

    def test_instance_of_different_methods(self):
        filepath = self.current_directory / "InstanceOfSampleDifferentMethods.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = InstanceOf()
        lines = pattern.value(ast)
        self.assertEqual(len(lines), 2)

    def test_instance_different_methods(self):
        filepath = self.current_directory / "InstanceSampleDifferentMethods.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = InstanceOf()
        lines = pattern.value(ast)
        self.assertEqual(len(lines), 2)

    def test_instance_of_several(self):
        filepath = self.current_directory / "InstanceOfSampleSeveral.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = InstanceOf()
        lines = pattern.value(ast)
        self.assertEqual(len(lines), 3)

    def test_instance_several(self):
        filepath = self.current_directory / "InstanceSampleSeveral.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = InstanceOf()
        lines = pattern.value(ast)
        self.assertEqual(len(lines), 2)

    def test_instance_in_method_chaining(self):
        filepath = self.current_directory / "InstanceSampleChain.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = InstanceOf()
        lines = pattern.value(ast)
        self.assertEqual(len(lines), 2)

    def test_both(self):
        filepath = self.current_directory / "InstanceBoth.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = InstanceOf()
        lines = pattern.value(ast)
        self.assertEqual(len(lines), 2)



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

from aibolit.ast_framework import AST, ASTNodeType


class NumberMethods:
    def value(self, ast: AST) -> int:
        metric = 0
        for class_node in ast.get_proxy_nodes(ASTNodeType.CLASS_DECLARATION):
            for method_node in class_node.methods:
                metric += 1
        return metric



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

from aibolit.ast_framework.computed_fields_registry import computed_fields_registry
from aibolit.ast_framework import ASTNodeType

from .nodes_filter import nodes_filter_factory
from .chained_fields import chain_field_getter_factory


def register_standard_computed_properties() -> None:
    _register_standard_nodes_filters()
    _register_standard_chain_fields()


def _register_standard_nodes_filters() -> None:
    computed_fields_registry.register(
        nodes_filter_factory("body", ASTNodeType.CONSTRUCTOR_DECLARATION),
        "constructors",
        ASTNodeType.CLASS_DECLARATION,
        ASTNodeType.INTERFACE_DECLARATION,
        ASTNodeType.ANNOTATION_DECLARATION,
    )

    computed_fields_registry.register(
        nodes_filter_factory("body", ASTNodeType.METHOD_DECLARATION),
        "methods",
        ASTNodeType.CLASS_DECLARATION,
        ASTNodeType.INTERFACE_DECLARATION,
        ASTNodeType.ANNOTATION_DECLARATION,
    )

    computed_fields_registry.register(
        nodes_filter_factory("body", ASTNodeType.FIELD_DECLARATION),
        "fields",
        ASTNodeType.CLASS_DECLARATION,
        ASTNodeType.INTERFACE_DECLARATION,
        ASTNodeType.ANNOTATION_DECLARATION,
    )

    computed_fields_registry.register(
        nodes_filter_factory("declarations", ASTNodeType.METHOD_DECLARATION),
        "methods",
        ASTNodeType.ENUM_DECLARATION,
    )

    computed_fields_registry.register(
        nodes_filter_factory("declarations", ASTNodeType.FIELD_DECLARATION),
        "fields",
        ASTNodeType.ENUM_DECLARATION,
    )


def _register_standard_chain_fields() -> None:
    computed_fields_registry.register(
        chain_field_getter_factory("declarators", "name"),
        "names",
        ASTNodeType.CONSTANT_DECLARATION,
        ASTNodeType.FIELD_DECLARATION,
        ASTNodeType.LOCAL_VARIABLE_DECLARATION,
        ASTNodeType.VARIABLE_DECLARATION,
    )



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

from typing import Dict

from cached_property import cached_property  # type: ignore
from deprecated import deprecated  # type: ignore

from aibolit.utils.ast_builder import build_ast
from aibolit.ast_framework import AST, ASTNodeType
from aibolit.ast_framework.java_class import JavaClass


@deprecated("This functionality must be transmitted to ASTNode")
class JavaPackage(AST):
    def __init__(self, filename: str):
        ast = AST.build_from_javalang(build_ast(filename))
        super().__init__(ast.tree, ast.root)

    @cached_property
    def name(self) -> str:
        try:
            package_declaration = next(
                self.children_with_type(self.root, ASTNodeType.PACKAGE_DECLARATION)
            )
            package_name = next(self.children_with_type(package_declaration, ASTNodeType.STRING))
            return self.tree.nodes[package_name]['string']
        except StopIteration:
            pass

        return '.'  # default package name

    @cached_property
    def java_classes(self) -> Dict[str, JavaClass]:
        classes: Dict[str, JavaClass] = {}
        for class_ast in self.get_subtrees(ASTNodeType.CLASS_DECLARATION):
            java_class = JavaClass(class_ast.tree, class_ast.root, self)
            classes[java_class.name] = java_class
        return classes



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT
from typing import List
from aibolit.ast_framework import ASTNodeType, AST
from aibolit.ast_framework.ast_node import ASTNode


class ClassicGetter:
    '''
    The method's name starts with get. Return statement is the only statement,
    excepting asserts.
    '''

    def _check_body_nodes(self, check_setter_body: List[ASTNode]) -> bool:
        '''
        Check whether nodes are agree with the following types
        (in self.suitable_nodes) or not.
        '''
        for node in check_setter_body:
            has_expression_value = hasattr(node, "expression")
            if not has_expression_value:
                return False

            is_return = node.node_type == ASTNodeType.RETURN_STATEMENT
            is_expression_memeber_ref = node.expression.node_type == ASTNodeType.MEMBER_REFERENCE
            if is_return and is_expression_memeber_ref:
                return True

        return False

    def value(self, ast: AST) -> List[int]:
        lines: List[int] = []
        for node in ast.get_proxy_nodes(ASTNodeType.METHOD_DECLARATION):
            method_name = node.name
            if method_name.startswith('get') and self._check_body_nodes(node.body):
                lines.append(node.line)
        return sorted(lines)



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

from typing import Set

from aibolit.ast_framework import AST, ASTNode, ASTNodeType


class FanOut:
    """
    Fan Out metric is defined as the number of other classes referenced by a class.
    """

    def value(self, ast: AST) -> int:
        fan_out = 0
        for class_declaration in ast.get_subtrees(ASTNodeType.CLASS_DECLARATION):
            fan_out += self._calculate_class_fan_out(class_declaration)

        return fan_out

    def _calculate_class_fan_out(self, java_class: AST) -> int:
        class_declaration = java_class.get_root()
        assert class_declaration.node_type == ASTNodeType.CLASS_DECLARATION

        used_classes_names: Set[str] = set()

        for type_reference in java_class.get_proxy_nodes(ASTNodeType.REFERENCE_TYPE):
            used_class_name = self._get_class_name_from_type_reference(type_reference)
            if used_class_name not in FanOut._excluded_class_names:
                used_classes_names.add(used_class_name)

        # remove name of the class
        used_classes_names -= {class_declaration.name}
        return len(used_classes_names)

    def _get_class_name_from_type_reference(self, type_reference: ASTNode) -> str:
        assert type_reference.node_type == ASTNodeType.REFERENCE_TYPE

        # type_reference 'name' field may not have a name of class in case this class
        # is referenced from packages, like 'package1.package2.class'.
        # To get class name we need to iterate over subtypes first and in the end get class name.
        while isinstance(type_reference.sub_type, ASTNode):
            type_reference = type_reference.sub_type

        return type_reference.name

    # exception are used from checkstyle.sourceforge.io/config_metrics.html#ClassFanOutComplexity
    # basic types ('int', 'long', etc) not used - ASTNodeType.REFERENCE_TYPE match only classes
    _excluded_class_names = {
        "ArrayIndexOutOfBoundsException",
        "ArrayList",
        "Boolean",
        "Byte",
        "Character",
        "Class",
        "Deprecated",
        "Deque",
        "Double",
        "Exception",
        "Float",
        "FunctionalInterface",
        "HashMap",
        "HashSet",
        "IllegalArgumentException",
        "IllegalStateException",
        "IndexOutOfBoundsException",
        "Integer",
        "LinkedList",
        "List",
        "Long",
        "Map",
        "NullPointerException",
        "Object",
        "Override",
        "Queue",
        "RuntimeException",
        "SafeVarargs",
        "SecurityException",
        "Set",
        "Short",
        "SortedMap",
        "SortedSet",
        "String",
        "StringBuffer",
        "StringBuilder",
        "SuppressWarnings",
        "Throwable",
        "TreeMap",
        "TreeSet",
        "UnsupportedOperationException",
        "Void",
        "System.out",
    }



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

from typing import List

from aibolit.ast_framework import AST, ASTNode, ASTNodeType


class JoinedValidation:
    '''
    Finds all if statements, which "then" branch consist of a single throw statement
    and logical or "||" used in condition.
    '''

    def value(self, ast: AST) -> List[int]:
        lines: List[int] = []
        for if_statement in ast.get_proxy_nodes(ASTNodeType.IF_STATEMENT):
            if self._is_logical_or_used_in_expression(if_statement.condition) and \
               self._is_block_consist_of_single_throw(if_statement.then_statement):
                lines.append(if_statement.line)
        return lines

    def _is_logical_or_used_in_expression(self, expression: ASTNode) -> bool:
        if expression.node_type == ASTNodeType.BINARY_OPERATION and expression.operator == '||':
            return True

        return any(self._is_logical_or_used_in_expression(child) for child in expression.children)

    def _is_block_consist_of_single_throw(self, block: ASTNode) -> bool:
        if block.node_type == ASTNodeType.THROW_STATEMENT:
            return True

        children = list(block.children)
        return len(children) == 1 and children[0].node_type == ASTNodeType.THROW_STATEMENT



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

from typing import List

from aibolit.ast_framework import AST, ASTNodeType


class ForceTypeCastingFinder:
    def value(self, ast: AST) -> List[int]:
        return [cast.expression.line for cast in ast.get_proxy_nodes(ASTNodeType.CAST)]



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

from aibolit.ast_framework import AST, ASTNode, ASTNodeType


class NCSSMetric:
    """
    NCSS metric counts non-commenting source statements.
    It counts:
     - keywords from _keyword_node_types
     - declarations from _declarations_node_types
     - local variable declarations and statement expressions
    """

    def value(self, ast: AST) -> int:
        metric = 0
        for node in ast.get_proxy_nodes(
            *NCSSMetric._keyword_node_types,
            *NCSSMetric._declarations_node_types,
            *NCSSMetric._misc_node_types
        ):
            metric += 1

            if node.node_type == ASTNodeType.IF_STATEMENT and self._has_pure_else_statements(node):
                metric += 1
            elif node.node_type == ASTNodeType.TRY_STATEMENT and self._has_finally_block(node):
                metric += 1

        return metric

    def _has_pure_else_statements(self, if_statement: ASTNode) -> bool:
        """
        Checks is there else branch.
        If else branch appeared to be "else if" construction (not pure "else"),
        returns False.
        """
        assert if_statement.node_type == ASTNodeType.IF_STATEMENT
        return if_statement.else_statement is not None and \
            if_statement.else_statement.node_type != ASTNodeType.IF_STATEMENT

    def _has_finally_block(self, try_statement: ASTNode) -> bool:
        assert try_statement.node_type == ASTNodeType.TRY_STATEMENT
        return try_statement.finally_block is not None

    # Two keywords "else" and "finally" are not represented by any nodes
    # and have to be extracted manually from fields
    # of IF_STATEMENT and TRY_STATEMENT respectively
    _keyword_node_types = {
        ASTNodeType.ASSERT_STATEMENT,
        ASTNodeType.BREAK_STATEMENT,
        ASTNodeType.CATCH_CLAUSE,
        ASTNodeType.CONTINUE_STATEMENT,
        ASTNodeType.DO_STATEMENT,
        ASTNodeType.FOR_STATEMENT,
        ASTNodeType.IF_STATEMENT,
        ASTNodeType.RETURN_STATEMENT,
        ASTNodeType.SWITCH_STATEMENT_CASE,
        ASTNodeType.SWITCH_STATEMENT,
        ASTNodeType.SYNCHRONIZED_STATEMENT,
        ASTNodeType.THROW_STATEMENT,
        ASTNodeType.TRY_STATEMENT,
        ASTNodeType.WHILE_STATEMENT,
    }

    # There are two type of declarations: type declarations (annotation, class etc.)
    # and class parts declarations (fields, methods etc.)
    # Package declarations are out of our interest, because it makes harder to compare
    # nested and outer classes.
    _declarations_node_types = {
        ASTNodeType.ANNOTATION_DECLARATION,
        ASTNodeType.CLASS_DECLARATION,
        ASTNodeType.CONSTANT_DECLARATION,
        ASTNodeType.CONSTRUCTOR_DECLARATION,
        ASTNodeType.ENUM_CONSTANT_DECLARATION,
        ASTNodeType.ENUM_DECLARATION,
        ASTNodeType.FIELD_DECLARATION,
        ASTNodeType.INTERFACE_DECLARATION,
        ASTNodeType.METHOD_DECLARATION,
        ASTNodeType.TYPE_DECLARATION,
    }

    _misc_node_types = {
        # Represent declaration of all local variables
        # *excluding* declarations inside "for" loop control.
        # Those declarations are represented by "VARIABLE_DECLARATION" nodes.
        ASTNodeType.LOCAL_VARIABLE_DECLARATION,
        # Statement expressions also includes assignments
        ASTNodeType.STATEMENT_EXPRESSION,
    }



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

from pathlib import Path
from unittest import TestCase

from aibolit.patterns.var_middle.var_middle import VarMiddle
from aibolit.ast_framework import AST
from aibolit.utils.ast_builder import build_ast


class VarMiddleTestCase(TestCase):
    current_directory = Path(__file__).absolute().parent

    def test_good_class(self):
        filepath = self.current_directory / "1.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = VarMiddle()
        lines = pattern.value(ast)
        self.assertEqual(lines, [])

    def test_bad_class(self):
        filepath = self.current_directory / "2.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = VarMiddle()
        lines = pattern.value(ast)
        self.assertEqual(lines, [12, 19])

    def test_case_with_multiline_method_declaration(self):
        filepath = self.current_directory / "3.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = VarMiddle()
        lines = pattern.value(ast)
        self.assertEqual(lines, [])

    def test_case_with_empty_lines(self):
        filepath = self.current_directory / "4.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = VarMiddle()
        lines = pattern.value(ast)
        self.assertEqual(lines, [])

    def test_case_autoclosable(self):
        filepath = self.current_directory / "5.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = VarMiddle()
        lines = pattern.value(ast)
        self.assertEqual(lines, [])

    def test_case_nested_class(self):
        filepath = self.current_directory / "6.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = VarMiddle()
        lines = pattern.value(ast)
        self.assertEqual(lines, [33, 36])

    def test_declaration_after_super_class_method_call(self):
        filepath = self.current_directory / "7.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = VarMiddle()
        lines = pattern.value(ast)
        self.assertEqual(lines, [17])

    def test_for_scope_good(self):
        filepath = self.current_directory / "8.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = VarMiddle()
        lines = pattern.value(ast)
        self.assertEqual(lines, [])

    def test_for_scope_bad(self):
        filepath = self.current_directory / "9.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = VarMiddle()
        lines = pattern.value(ast)
        self.assertEqual(lines, [14])

    def test_variable_declared_after_for(self):
        filepath = self.current_directory / "10.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = VarMiddle()
        lines = pattern.value(ast)
        self.assertEqual(lines, [14])

    def test_11(self):
        filepath = self.current_directory / "11.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = VarMiddle()
        lines = pattern.value(ast)
        self.assertEqual(lines, [])

    def test_catch_good(self):
        filepath = self.current_directory / "12.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = VarMiddle()
        lines = pattern.value(ast)
        self.assertEqual(lines, [])

    def test_catch_bad(self):
        filepath = self.current_directory / "13.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = VarMiddle()
        lines = pattern.value(ast)
        self.assertEqual(lines, [41])

    def test_else_bad(self):
        filepath = self.current_directory / "14.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = VarMiddle()
        lines = pattern.value(ast)
        self.assertEqual(lines, [90])

    def test_variable_after_curly_braces(self):
        filepath = self.current_directory / "15.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = VarMiddle()
        lines = pattern.value(ast)
        self.assertEqual(lines, [])

    def test_variable_inside_lambda(self):
        filepath = self.current_directory / "16.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = VarMiddle()
        lines = pattern.value(ast)
        self.assertEqual(lines, [])

    def test_annotation_with_parameters(self):
        filepath = self.current_directory / "17.java"
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = VarMiddle()
        lines = pattern.value(ast)
        self.assertEqual(lines, [25])



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

from unittest import TestCase
from pathlib import Path
from typing import List

from aibolit.patterns.partially_synchronized_methods.partially_synchronized_methods import (
    PartiallySynchronizedMethods,
)
from aibolit.ast_framework import AST
from aibolit.utils.ast_builder import build_ast


class PartiallySynchronizedMethodsTestCase(TestCase):
    def test_no_synchronization(self):
        self._test_helper("NoSynchronization.java", [])

    def test_partial_synchronization(self):
        self._test_helper("PartialSynchronization.java", [8, 16, 26])

    def test_full_synchronization(self):
        self._test_helper("FullSynchronization.java", [])

    def test_several_synchronization_statements(self):
        self._test_helper("SeveralSynchronizationStatements.java", [9, 13, 29, 32])

    def test_synchronization_in_nested_scopes(self):
        self._test_helper("SynchronizationInNestedScopes.java", [12, 21, 30])

    def _test_helper(self, filename: str, lines: List[int]):
        filepath = str(Path(__file__).absolute().parent / filename)
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = PartiallySynchronizedMethods()
        self.assertEqual(pattern.value(ast), lines)



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

from unittest import TestCase
from pathlib import Path

from aibolit.ast_framework import AST, ASTNodeType
from aibolit.utils.ast_builder import build_ast


class ASTNodeTestSuite(TestCase):
    def test_class_computed_fields(self):
        ast = AST.build_from_javalang(
            build_ast(
                Path(__file__).absolute().parent / "MethodUseOtherMethodExample.java"
            )
        )
        package = ast.get_root()
        assert len(package.types) == 1 and \
            package.types[0].node_type == ASTNodeType.CLASS_DECLARATION

        java_class = package.types[0]
        self.assertEqual(java_class.name, "MethodUseOtherMethod")
        self.assertEqual(java_class.modifiers, set())
        self.assertEqual(java_class.documentation, "/**\n* Some documentation\n*/")

        # consider each field declaration declares single field
        fields_names = {field.names[0] for field in java_class.fields}
        self.assertEqual(fields_names, {"connectingField", "redundantField"})

        methods_names = {method.name for method in java_class.methods}
        self.assertEqual(
            methods_names,
            {
                "useOnlyMethods1",
                "useOnlyMethods2",
                "getField",
                "setField",
                "standAloneMethod",
                "shadowing",
            },
        )

        self.assertEqual(set(java_class.constructors), set())



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

from pathlib import Path
from unittest import TestCase

from aibolit.patterns.hybrid_constructor.hybrid_constructor import HybridConstructor
from aibolit.ast_framework import AST
from aibolit.utils.ast_builder import build_ast


class HybridConstructorTestCase(TestCase):
    cur_dir = Path(__file__).absolute().parent

    def test_several(self):
        filepath = Path(self.cur_dir, "several.java")
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = HybridConstructor()
        lines = pattern.value(ast)
        self.assertEqual(lines, [7, 13, 23])

    def test_simple2(self):
        filepath = Path(self.cur_dir, "init_block.java")
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = HybridConstructor()
        lines = pattern.value(ast)
        self.assertEqual(lines, [])

    def test_simple22(self):
        filepath = Path(self.cur_dir, "init_static_block.java")
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = HybridConstructor()
        lines = pattern.value(ast)
        self.assertEqual(lines, [])

    def test_simple3(self):
        filepath = Path(self.cur_dir, "autocloseable.java")
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = HybridConstructor()
        lines = pattern.value(ast)
        self.assertEqual(lines, [7, 17, 34])

    def test_simple5(self):
        filepath = Path(self.cur_dir, "one_line_usage.java")
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = HybridConstructor()
        lines = pattern.value(ast)
        self.assertEqual(lines, [15])

    def test_simple6(self):
        filepath = Path(self.cur_dir, "super.java")
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = HybridConstructor()
        lines = pattern.value(ast)
        self.assertEqual(lines, [15])

    def test_simple7(self):
        filepath = Path(self.cur_dir, "holy_moly_constructor.java")
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = HybridConstructor()
        lines = pattern.value(ast)
        self.assertEqual(lines, [50])

    def test_simple9(self):
        filepath = Path(self.cur_dir, "super_this.java")
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = HybridConstructor()
        lines = pattern.value(ast)
        self.assertEqual(lines, [18, 28, 54, 65, 79, 90, 104])

    def test_simple10(self):
        filepath = Path(self.cur_dir, "BookmarkEditCmd.java")
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = HybridConstructor()
        lines = pattern.value(ast)
        self.assertEqual(lines, [])

    def test_simple11(self):
        filepath = Path(self.cur_dir, "ChainedBuffer.java")
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = HybridConstructor()
        lines = pattern.value(ast)
        self.assertEqual(lines, [])

    def test_simple12(self):
        filepath = Path(self.cur_dir, "CliMethodExtraSections.java")
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = HybridConstructor()
        lines = pattern.value(ast)
        self.assertEqual(lines, [])

    def test_simple13(self):
        filepath = Path(self.cur_dir, "LengthStringOrdinalSet.java")
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = HybridConstructor()
        lines = pattern.value(ast)
        self.assertEqual(lines, [])

    def test_simple14(self):
        filepath = Path(self.cur_dir, "LoaderInfoHeader.java")
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = HybridConstructor()
        lines = pattern.value(ast)
        self.assertEqual(lines, [])

    def test_simple15(self):
        filepath = Path(self.cur_dir, "OmfModuleEnd.java")
        ast = AST.build_from_javalang(build_ast(filepath))
        pattern = HybridConstructor()
        lines = pattern.value(ast)
        self.assertEqual(lines, [])



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT

from typing import List, Dict, Set, Iterator, Any

from networkx import (  # type: ignore
    DiGraph, strongly_connected_components, weakly_connected_components
)

from aibolit.ast_framework import AST, ASTNodeType

from aibolit.patterns.classic_setter.classic_setter import ClassicSetter as setter
from aibolit.patterns.classic_getter.classic_getter import ClassicGetter as getter


def find_patterns(tree: AST, patterns: List[Any]) -> Set[str]:
    """
    Searches all setters in a component
    :param patterns: list of patterns to check
    :param tree: ast tree
    :return: list of method name which are setters
    """

    patterns_method_names: Set[str] = set()
    for method_declaration in tree.get_root().methods:
        method_ast = tree.get_subtree(method_declaration)
        for pattern in patterns:
            if is_ast_pattern(method_ast, pattern):
                patterns_method_names.add(method_declaration.name)

    return patterns_method_names


def is_ast_pattern(class_ast: AST, Pattern) -> bool:
    """
    Checks whether ast is some pattern
    :param Pattern: pattern class
    :param class_ast: ast tree
    :return: True if it is setter, otherwise - False
    """
    return len(Pattern().value(class_ast)) > 0


def decompose_java_class(
        class_ast: AST,
        strength: str,
        ignore_setters=False,
        ignore_getters=False) -> List[AST]:
    """
    Splits java_class fields and methods by their usage and
    construct for each case an AST with only those fields and methods kept.
    :param class_ast: component
    :param ignore_getters: should ignore getters
    :param ignore_setters: should ignore setters
    :param strength: controls splitting criteria. Use "strong" or "weak"
    for splitting fields and methods by strong and weak connectivity.
    """

    usage_graph = _create_usage_graph(class_ast)

    components: Iterator[Set[int]]
    if strength == "strong":
        components = strongly_connected_components(usage_graph)
    elif strength == "weak":
        components = weakly_connected_components(usage_graph)
    else:
        raise ValueError(
            f"'strength' argument must be either 'strong' or 'weak', but '{strength}' was provided."
        )

    class_parts: List[AST] = []
    patterns_to_ignore: List[Any] = []
    if ignore_getters:
        patterns_to_ignore.append(getter)
    if ignore_setters:
        patterns_to_ignore.append(setter)

    if ignore_setters or ignore_getters:
        prohibited_function_names = find_patterns(class_ast, patterns_to_ignore)

    for component in components:

        field_names = {
            usage_graph.nodes[node]["name"]
            for node in component
            if usage_graph.nodes[node]["type"] == "field"
        }
        method_names = {
            usage_graph.nodes[node]["name"]
            for node in component
            if usage_graph.nodes[node]["type"] == "method"
        }

        if ignore_setters or ignore_getters:
            method_names = method_names.difference(prohibited_function_names)

        filtered = _filter_class_methods_and_fields(class_ast, field_names, method_names)

        class_parts.append(
            filtered
        )

    return class_parts


def _create_usage_graph(class_ast: AST) -> DiGraph:
    usage_graph = DiGraph()
    fields_ids: Dict[str, int] = {}
    methods_ids: Dict[str, int] = {}

    class_declaration = class_ast.get_root()

    for field_declaration in class_declaration.fields:
        # several fields can be declared at one line
        for field_name in field_declaration.names:
            fields_ids[field_name] = len(fields_ids)
            usage_graph.add_node(fields_ids[field_name], type="field", name=field_name)

    for method_declaration in class_declaration.methods:
        method_name = method_declaration.name

        # overloaded methods considered as single node in usage_graph
        if method_name not in methods_ids:
            methods_ids[method_name] = len(fields_ids) + 1 + len(methods_ids)
            usage_graph.add_node(
                methods_ids[method_name], type="method", name=method_name
            )

    for method_declaration in class_declaration.methods:
        method_ast = class_ast.get_subtree(method_declaration)

        for invoked_method_name in _find_local_method_invocations(method_ast):
            if invoked_method_name in methods_ids:
                usage_graph.add_edge(
                    methods_ids[method_declaration.name],
                    methods_ids[invoked_method_name],
                )

        for used_field_name in _find_fields_usage(method_ast):
            if used_field_name in fields_ids:
                usage_graph.add_edge(
                    methods_ids[method_declaration.name], fields_ids[used_field_name]
                )

    return usage_graph


def _find_local_method_invocations(method_ast: AST) -> Set[str]:
    invoked_methods: Set[str] = set()
    for method_invocation in method_ast.get_proxy_nodes(ASTNodeType.METHOD_INVOCATION):
        if method_invocation.qualifier is None:
            invoked_methods.add(method_invocation.member)

    return invoked_methods


def _find_fields_usage(method_ast: AST) -> Set[str]:
    local_variables: Set[str] = set()
    for variable_declaration in method_ast.get_proxy_nodes(
            ASTNodeType.LOCAL_VARIABLE_DECLARATION
    ):
        local_variables.update(variable_declaration.names)

    method_declaration = method_ast.get_root()
    for parameter in method_declaration.parameters:
        local_variables.add(parameter.name)

    used_fields: Set[str] = set()
    for member_reference in method_ast.get_proxy_nodes(ASTNodeType.MEMBER_REFERENCE):
        if member_reference.qualifier is None and \
                member_reference.member not in local_variables:
            used_fields.add(member_reference.member)

    return used_fields


def _filter_class_methods_and_fields(
        class_ast: AST,
        allowed_fields_names: Set[str],
        allowed_methods_names: Set[str]
) -> AST:
    class_declaration = class_ast.get_root()
    allowed_nodes = {class_declaration.node_index}

    for field_declaration in class_declaration.fields:
        if len(allowed_fields_names & set(field_declaration.names)) != 0:
            field_ast = class_ast.get_subtree(field_declaration)
            allowed_nodes.update(node.node_index for node in field_ast)

    for method_declaration in class_declaration.methods:
        if method_declaration.name in allowed_methods_names:
            method_ast = class_ast.get_subtree(method_declaration)
            allowed_nodes.update(node.node_index for node in method_ast)

    return AST(class_ast.tree.subgraph(allowed_nodes), class_declaration.node_index)



# SPDX-FileCopyrightText: Copyright (c) 2019-2025 Aibolit
# SPDX-License-Identifier: MIT
from typing import List

from aibolit.ast_framework import ASTNodeType, AST
from aibolit.ast_framework.ast_node import ASTNode


class PublicStaticMethod:
    '''
    Check method with 'public static'
    '''
    def _check_public_static(self, node: ASTNode):
        if all(type in node.modifiers for type in ['public', 'static']):
            return True
        return False

    def value(self, ast: AST):
        lines: List[int] = []
        for method_declaration in ast.get_proxy_nodes(ASTNodeType.METHOD_DECLARATION):
            if self._check_public_static(method_declaration):
                lines.append(method_declaration.line)
        return lines
